# -*- coding: iso-8859-1 -*-
# $Id$
# Snowberry: Extensible Launcher for the Doomsday Engine
#
# Copyright (C) 2004, 2006
#   Jaakko Keränen <jaakko.keranen@iki.fi>
#   Veikko Eeva
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not: http://www.opensource.org/

## @file addon.py Addon Classes

import os, re, string, zipfile, shutil, struct, traceback
import logger, paths, language, cfparser
import aodb, confdb


EXCLUDES = 0
REQUIRES = 1
PROVIDES = 2
OFFERS = 3

# File name extensions of the supported addon formats.
EXTENSIONS = ['box', 'addon', 'pk3', 'zip', 'lmp', 'wad', 'ded', 'deh']

# All the possible names of the addon metadata file.
META_NAMES = ['Info', 'Info.conf', 'info', 'info.conf']
    

class Addon:
    """Addon class serves as an abstract base class to the concrete
    addon implementations."""

    def __init__(self, identifier, source):
        """Construct a new Addon object.

        @param identifier The identifier of the addon.

        @param source The full path name of the source file.
        """
        self.id = identifier
        self.name = ''
        self.category = aodb.getRootCategory()
        self.source = source
        self.priority = 'z'
        self.requiredComponents = []
        self.excludedCategories = []
        self.box = None

        # When an addon is uninstalled, it will be hidden in the user
        # interface. 
        self.uninstalled = False

        # Contains tuples: (type, identifier).
        self.keywords = {}
        for t in [EXCLUDES, REQUIRES, PROVIDES, OFFERS]:
            self.keywords[t] = []

        # Every addon provides itself.
        self.keywords[PROVIDES].append(identifier)
        
        self.lastModified = None

        # If no other translation is given, the identifier of the
        # addon is shown as the base name of the source.
        if not language.isDefined(identifier):
            language.define('english', identifier, os.path.basename(source))

    def getId(self):
        return self.id

    def getType(self):
        """Returns the type identifier of the addon class."""

        return 'addon-type-generic'

    def getName(self):
        return self.name

    def setName(self, name):
        self.name = name

    def getSource(self):
        """The source path is the location of the addon.

        @return The source path of the addon as a string.
        """
        return self.source

    def getCategory(self):
        return self.category

    def getExcludedCategories(self):
        return self.excludedCategories

    def makeLocalId(self, identifier):
        return self.id + '-' + identifier

    def getContentPath(self):
        """The content path is the location on the file system where
        the contents of the addon are stored.  This is not meaningful
        with all addons.
        """
        return self.source

    def getLastModified(self):
        """Returns the time when the addon data was last modified.
        This is used to determine if the metadata needs to be updated.
        """
        if self.isUninstalled():
            # Uninstalled addons don't have a meaningful last modified
            # timestamp.
            return 0
        
        if self.lastModified == None:
            self.lastModified = self.determineLastModified()

        return self.lastModified

    def determineLastModified(self):
        """Read the source file(s) to see when they've been modified
        last.  This implementation assumes the source is a single
        file.

        @return The last modification time.
        """
        try:
            return os.path.getmtime(self.source)
        except OSError:
            return 0

    def setBox(self, box):
        """Set the BoxAddon this addon is inside of.

        @param box A BoxAddon object."""

        self.box = box

    def getBox(self):
        """Returns the box of this addon.

        @param BoxAddon object.  May be None.
        """

        return self.box

    def isBox(self):
        return False

    def isUninstalled(self):
        return self.uninstalled

    def uninstall(self):
        """Mark this addon as uninstalled.  It will be hidden in the
        user interface."""
        
        self.uninstalled = True

    def isVisible(self):
        """All visible addons are listen in the addons tree.  The
        required parts of a box are not visible, as they are loaded
        automatically if the box itself is selected for loading."""

        if self.isUninstalled():
            # Uninstalled addons are not visible.
            return False

        if not self.getBox():
            return True

        return not self.getBox().isRequired(self.getId())

    def isOptIn(self, contextProfile):
        """Opt-in addons must be attached to a profile to become used.

        @param contextProfile The profiles.Profile object that will be
        used as the context.
        """
        import sb.profdb
        defaults = sb.profdb.getDefaults()

        optIn = True

        if contextProfile is not defaults:
            if self.getId() in defaults.getAddons():
                # This addon is attached to the Defaults.
                optIn = not optIn

        if self.isInversed():
            optIn = not optIn

        return optIn

    def isOptOut(self, contextProfile):
        """Opt-out addons must be detached from a profile to become
        used.

        @param contextProfile The profiles.Profile object that will be
        used as the context.
        """
        
        return not self.isOptIn(self, contextProfile)

    def isInversed(self):
        """Inversed addons are used when not attached, and not used
        when attached to a profile.  For example: optional parts of a
        box."""

        if not self.getBox():
            return False

        # Optional parts of boxes.
        if self.getBox().isOptional(self.getId()):
            return True

        return False

    def setPriority(self, priorityClass):
        """The priority class is an ASCII character.

        @param priorityClass Expected to be a character between A and
        Z (inclusive).
        """
        self.priority = priorityClass.lower()

    def getPriority(self):
        return self.priority

    def readMetaData(self):
        """Read the source file(s) and read/generate the metadata for
        the addon.  The data is stored in the Addon object.  This
        implementation in the base class handles the data is common to
        all addons.
        """
        pass

    def getCommandLine(self, profile):
        """Return the command line options to load and configure the
        addon.  This basic implementation assumes no configurability
        and a single source file.
        """
        if not paths.hasExtension('manifest', self.source):
            return '-file ' + paths.quote(self.source)
        else:
            return ''

    def addKeywords(self, type, identifiers):
        """Define a new keyword for the addon.

        @param type One of the constants:
        - EXCLUDES: Identifier must not be used by anyone else.
        - REQUIRES: Identifier must be used by someone else.
        - PROVIDES: Identifier is provided by this addon.
        - OFFERS: Identifier is provided, unless a higher priority
          addon provides or offers the identifier.

        @param identifiers An array of identifiers.  These identifiers
        live in their own namespace, but the values of settings are
        also considered keywords.
        """
        self.keywords[type] += identifiers

    def getKeywords(self, type):
        return self.keywords[type]    

    def isCompatibleWith(self, profile):
        """Tests if this addon is compatible with the profile.

        @return Boolean.  True, if compatible.
        """
        if self.isUninstalled():
            return False
        
        # The Defaults profile is compatible with everything.
        import sb.profdb
        if profile is sb.profdb.getDefaults():
            return True

        # Check the required components.  The profile must have all of
        # them.
        profComps = profile.getComponents()
        for req in self.requiredComponents:
            if req not in profComps:
                return False

        return True

    def parseConfiguration(self, configSource):
        """Parse the given configuration data and process it as this
        addon's metadata.

        @param configSource The raw configuration data.
        """
        try:
            p = cfparser.Parser(configSource)

            while True:
                elem = p.get()

                if elem.isBlock() and elem.getType() == 'language':
                    language.processLanguageBlock(elem, self.getId())

                elif elem.isKey() and elem.getName() == 'name':
                    self.setName(elem.getValue())
                    
                    # A shortcut for defining the name of the addon.
                    language.define('english', self.getId(), self.getName())
                    
                    # Automatically define the name of a box part.
                    if self.getBox():
                        language.define('english',
                                        self.getBox().getId() + '-' +
                                        self.getId(), self.getName())
                    
                elif elem.getName() == 'component':
                    self.requiredComponents += elem.getContents()

                elif elem.isKey() and elem.getName() == 'category':
                    # The category of an addon inside a box is
                    # relative to the box.
                    if self.getBox():
                        catPath = self.getBox().getCategory().getPath() + \
                                  '/' + self.getBox().getId() + '/'
                    else:
                        catPath = ''
                    catPath += elem.getValue()
                    self.category = aodb._getCategory(catPath)

                elif elem.isKey() and elem.getName() == 'order':
                    self.setPriority(elem.getValue())

                elif elem.getName() == 'excludes-category':
                    identifiers = elem.getContents()
                    for id in identifiers:
                        self.excludedCategories.append(aodb._getCategory(id))

                elif elem.getName() == 'excludes':
                    self.addKeywords(EXCLUDES, elem.getContents())

                elif elem.getName() == 'requires':
                    self.addKeywords(REQUIRES, elem.getContents())
                    
                elif elem.getName() == 'provides':
                    self.addKeywords(PROVIDES, elem.getContents())

                elif elem.getName() == 'offers':
                    self.addKeywords(OFFERS, elem.getContents())

                elif elem.isBlock():
                    # Convert the identifier to local form.
                    elem.setName(self.makeLocalId(elem.getName()))

                    # Determine the owner of this setting.
                    if self.getBox() == None:
                        ownerId = self.getId()
                    else:
                        ownerId = self.getBox().getId()
                        # The addon's identifier determines the group.
                        elem.add(cfparser.KeyElement('group', self.getId()))
                    elem.add(cfparser.KeyElement('addon', ownerId))
                    import confdb
                    confdb.processSettingBlock(elem)

        except cfparser.OutOfElements:
            pass


class BoxAddon (Addon):
    """A Box groups other addons together.  A box can't contain
    another box."""

    def __init__(self, identifier, source):
        Addon.__init__(self, identifier, source)

        # Register the source directory into the paths.
        paths.addBundlePath(self.getContentPath())

        # Lists of addon identifiers.  These are the parts of the box.
        self.required = []
        self.optional = []
        self.extra = []

    def getType(self):
        return 'addon-type-box'

    def isBox(self):
        return True

    def isRequired(self, addonIdentifier):
        return addonIdentifier in self.required

    def isOptional(self, addonIdentifier):
        return addonIdentifier in self.optional

    def getRequiredParts(self):
        return self.required

    def getContentCategoryLongId(self):
        """Forms the category long identifier, in which the contents of this box 
        will be under.
        @return Category long id.
        """
        return self.getCategory().getLongId() + '-' + self.getId()

    def __makePath(self, name):
        return os.path.join(self.getContentPath(), name)

    def uninstall(self):
        """Mark the box and its contents uninstalled."""

        Addon.uninstall(self)

        for ident in self.required + self.optional + self.extra:
            get(ident).uninstall()

    def loadContents(self):
        """Load all the addons inside the box."""
        
        self.loadAll(self.getContentPath(), self.optional)
        self.loadAll(self.__makePath('Required'), self.required)
        self.loadAll(self.__makePath('Extra'), self.extra)
        if paths.isCaseSensitive():
            self.loadAll(self.__makePath('required'), self.required)
            self.loadAll(self.__makePath('extra'), self.extra)

    def loadAll(self, path, role):
        """Load all addons on the given path."""

        if not os.path.exists(path):
            return

        for fileName in os.listdir(path):
            try:
                ident = aodb.load(os.path.join(path, fileName), self)
                role.append(ident)

                # Using any parts of an addon requires that the box
                # itself is marked for loading.
                aodb.get(ident).addKeywords(REQUIRES, [self.getId()])

            except aodb.LoadError, x:
                # Attempted to load a file that is not an addon.
                # Just ignore; boxes can contain all kinds of stuff.
                pass

            except Exception, x:
                # Perhaps it wasn't an addon?
                traceback.print_exc()
            
    def readMetaData(self):
        """Read the metadata files (e.g. Info)."""

        try:
            for info in META_NAMES:
                try:
                    name = os.path.join(self.getContentPath(), info)
                    conf = file(name).read()
                    self.parseConfiguration(conf)
                except OSError:
                    pass
                except IOError:
                    pass
                except:
                    logger.add(logger.HIGH, 'error-read-info-file',
                               name, self.getId())
        except:
            traceback.print_exc()
 
        # Load a readme from a separate file.
        readme = self.__makePath('Readme.html')
        if not os.path.exists(readme):
            readme = self.__makePath('readme.html')
        if os.path.exists(readme):
            language.define('english',
                            self.makeLocalId('readme'),
                            file(readme).read())
                            
    def getCommandLine(self, profile):
        """Returns the command line that is given to Doomsday Engine.

        @param profile The profile for which the command line is being
        generated.
        @return String The generated command line.
        """
        return ""
                            

class BundleAddon (Addon):
    """The Bundle is the most sophicated of the addons.  It is a
    directory that follows a specific struture."""

    def __init__(self, identifier, source):
        Addon.__init__(self, identifier, source)

        # Register the source directory into the paths.
        paths.addBundlePath(self.getContentPath())

    def getType(self):
        return 'addon-type-bundle'

    def determineLastModified(self):
        """Read the source file(s) to see when they've been modified
        last.  This implementation assumes the source is a single
        file.

        @return The last modification time.
        """
        return aodb._getLatestModTime(self.source)

    def __makePath(self, fileName):
        """Make a path name for a file inside the bundle."""
        return os.path.join(self.source, 'Contents', fileName)

    def getContentPath(self):
        """The content path of a bundle is well-defined.
        """
        return os.path.join(self.source, 'Contents')

    def readMetaData(self):
        """Read the metadata files (e.g. Contents/Info)."""

        try:
            for info in META_NAMES:
                try:
                    conf = file(self.__makePath(info)).read()
                    self.parseConfiguration(conf)
                except OSError:
                    pass
                except IOError:
                    pass
                except Exception, x:
                    logger.add(logger.HIGH, 'error-read-info-file',
                               self.__makePath(info), self.getId())
        except:
            # Log a warning?
            traceback.print_exc()

        # Load a readme from a separate file.
        readme = self.__makePath('Readme.html')
        if not os.path.exists(readme):
            readme = self.__makePath('readme.html')
        if os.path.exists(readme):
            language.define('english',
                            self.makeLocalId('readme'),
                            file(readme).read())

    def getCommandLine(self, profile):
        """Bundles are loaded with -file and -def.  We'll load the
        contents of the Defs and Data directories.
        """
        param = ''

        defsPath = self.__makePath('Defs')
        if os.path.exists(defsPath):
            defs = []

            # Only include the files that have the .ded extension.
            for name in os.listdir(defsPath):
                path = os.path.join(defsPath, name)
                if paths.hasExtension('ded', path):
                    defs.append(paths.quote(path))

            if len(defs) > 0:
                # At least one definition file was found.
                param += '-def ' + string.join(defs)

        dataPath = self.__makePath('Data')
        if os.path.exists(dataPath):
            data = []

            # Only include the files with known extensions.
            for name in os.listdir(dataPath):
                path = os.path.join(dataPath, name)
                if paths.hasExtension('pk3', path) or \
                   paths.hasExtension('wad', path) or \
                   paths.hasExtension('lmp', path):
                    data.append(paths.quote(path))

            if len(data):
                # At least one data file was found.
                param += ' -file ' + string.join(data)

        return param


class PK3Addon (Addon):
    def __init__(self, identifier, source):
        Addon.__init__(self, identifier, source)

    def getType(self):
        return 'addon-type-pk3'

    def readMetaData(self):
        """Read the metadata file(s)."""

        # Check if there is an Info lump.
        try:
            z = zipfile.ZipFile(self.source)
            for info in META_NAMES:
                try:
                    self.parseConfiguration(z.read(info))
                except KeyError:
                    pass
                except:
                    logger.add(logger.HIGH, 'error-read-zip-info-file',
                               self.source, self.getId())

        except:
            # Log a warning?
            print "%s: Failed to open archive." % self.source


class WADAddon (Addon):
    def __init__(self, identifier, source):
        Addon.__init__(self, identifier, source)

        # We assume this is a PWAD until closer inspection.
        self.wadType = 'PWAD'
        self.lumps = []

    def getType(self):
        return 'addon-type-wad'
        
    def isPWAD(self):   
        """Determines if this addon is a PWAD addon."""
        return self.wadType == 'PWAD'
        
    def getCommandLine(self, profile):
        # The Deathkings WAD is a special case and will get always loaded
        # with the -iwad option.
        if self.getId() == 'hexdd-wad':
            return '-iwad ' + paths.quote(self.source)
        return Addon.getCommandLine(self, profile)

    def readMetaData(self):
        """Generate metadata by making guesses based on the WAD file
        name and contents."""

        metadata = ''

        # Defaults.
        game = ''

        # Look at the path where the WAD file is located.
        path = self.getContentPath().lower()
        # But ignore the user addon path.
        if path.find(paths.getUserPath(paths.ADDONS).lower()) == 0:
            path = path[len(paths.getUserPath(paths.ADDONS)) + 1:]

        if 'heretic' in path or 'htic' in path:
            game = 'jheretic'

        elif 'hexen' in path or 'hxn' in path or 'hexendk' in path or \
             'deathkings' in path:
            game = 'jhexen'

        elif 'doom' in path or 'doom2' in path or 'final' in path or \
             'finaldoom' in path or 'tnt' in path or 'plutonia' in path:
            game = 'jdoom'

        # Check for a game component name in the path.
        for component in confdb.getGameComponents():
            compName = component.getId()[5:] # skip "game-" prefix
            if compName in path:
                game = compName
                break

        # Read the WAD directory.
        self.lumps = self.readLumpDirectory()
        lumps = self.lumps
        
        if 'BEHAVIOR' in lumps:
            # Hexen-specific lumps.
            game = 'jhexen'

        if 'ADVISOR' in lumps or 'M_HTIC' in lumps:
            game = 'jheretic'

        if 'M_DOOM' in lumps:
            game = 'jdoom'

        # What can be determined by looking at the lump names?
        for lump in lumps:
            if lump[:2] == 'D_':
                # Doom music lump.
                game = 'jdoom'

        episodic = False
        maps = False

        # Are there any ExMy maps in the WAD?
        for lump in lumps:
            if len(lump) == 4 and lump[0] == 'E' and lump[2] == 'M' and \
               lump[1] in string.digits and lump[3] in string.digits:
                episodic = True

                # Episodes 5 and 6 exist in Heretic.
                if game == '' and (lump[1] == '5' or lump[1] == '6'):
                    game = 'jheretic'

        # Are there any MAPxy in the WAD?
        for lump in lumps:
            if len(lump) == 5 and lump[:3] == 'MAP' and \
               lump[3] in string.digits and lump[4] in string.digits:
                maps = True
                break

        if episodic and game == 'jhexen':
            # Guessed wrong.
            game = ''

        if maps and game == 'jheretic':
            # Guessed wrong.
            game = ''

        # Determine a category for the WAD.
        if self.wadType == 'IWAD':
            metadata += "category: gamedata/primary\n"
        elif self.wadType == 'JWAD':
            metadata += "category: gamedata/doomsday\n"
        elif self.wadType == 'PWAD':
            category = 'gamedata/'

            if maps or episodic:
                category += 'maps/'

                # Category based on the name.	 
                base = paths.getBase(self.getContentPath()).lower()	 

                if base[0] in string.digits:	 
                    category += '09/'	 
                if base[0] in 'abcdefg':	 
                    category += 'ag/'	 
                if base[0] in 'hijklm':	 
                    category += 'hm/'	 
                if base[0] in 'nopqrs':	 
                    category += 'ns/'	 
                if base[0] in 'tuvwxyz':	 
                    category += 'tz/'	 
             
            metadata += "category: %s\n" % category

        # Game component.
        if game != '':
            metadata += "component: game-%s\n" % game

        self.parseConfiguration(metadata)

    def getShortContentAnalysis(self):
        """Forms a short single line of text that describes the contents
        of the WAD."""

        # The directory must already be read by this time.
        analysis = []              
        mapCount = 0
        hasCustomTextures = False
        
        for lump in self.lumps:
            # Count the number of maps.
            if len(lump) == 4 and lump[0] == 'E' and lump[2] == 'M' and \
               lump[1] in string.digits and lump[3] in string.digits:
                mapCount += 1
            elif len(lump) == 5 and lump[:3] == 'MAP' and \
               lump[3] in string.digits and lump[4] in string.digits:
                mapCount += 1
            # Any custom textures?
            elif lump == 'PNAMES' or lump == 'TEXTURE1' or lump == 'TEXTURE2':
                hasCustomTextures = True
        
        if mapCount > 0:
            analysis.append( language.expand(
                language.translate('wad-analysis-maps'), str(mapCount)) )
        if hasCustomTextures:
            analysis.append(language.translate('wad-analysis-custom-textures'))
        
        return string.join(analysis, ', ').capitalize()

    def readLumpDirectory(self):
        """Open the WAD file and read the lump directory at the end of
        the file."""

        lumps = []

        try:
            f = file(self.getContentPath(), 'rb')

            # Type of the WAD file. This affects categorization.
            self.wadType = f.read(4)

            # WADs use little-endian byte ordering.
            count, dirOffset = struct.unpack('<II', f.read(8))
           
            f.seek(dirOffset)

            while count > 0:
                # Pos and size ignored for now.
                #filePos, lumpSize = struct.unpack('II', f.read(8))
                f.seek(8, 1)

                # Make sure the name is in upper case.
                name = f.read(8).upper()
                if len(name) != 8:
                    break
                
                for i in range(len(name)):
                    if ord(name[i]) < 32:
                        name = name[:i]
                        break

                if len(name) > 0:
                    lumps.append(name)

                count -= 1

            f.close()

        except Exception, x:
            raise x
            # Ignore errors.
            #pass

        return lumps

        
class DehackedAddon (Addon):
    def __init__(self, identifier, source):
        Addon.__init__(self, identifier, source)

    def getType(self):
        return 'addon-type-dehacked'

    def getCommandLine(self, profile):
        """Return the command line options to load and configure the
        addon.  Use <tt>-deh</tt> for loading DeHackEd files.
        
        @todo  Check profile for extra options?
        """
        return '-deh ' + paths.quote(self.source)

    def readMetaData(self):
        """Generate metadata by making guesses based on the DEH patch
        name and contents."""

        metadata = 'category: patches'

        self.parseConfiguration(metadata)
    
        

class DEDAddon (Addon):
    def __init__(self, identifier, source):
        Addon.__init__(self, identifier, source)

    def getType(self):
        return 'addon-type-ded'
        
    def getCommandLine(self, profile):
        """Return the command line options to load and configure the
        addon.  Use <tt>-def</tt> for loading definition files.
        
        @todo  Check profile for extra options?
        """
        return '-def ' + paths.quote(self.source)

    def readMetaData(self):
        """Generate metadata by making guesses based on the DED patch
        name and contents."""

        metadata = 'category: definitions'

        self.parseConfiguration(metadata)


class LumpAddon (Addon):
    def __init__(self, identifier, source):
        Addon.__init__(self, identifier, source)

    def getType(self):
        return 'addon-type-lump'
        
    def readMetaData(self):
        """Generate metadata for the lump."""

        metadata = 'category: gamedata/lumps'

        self.parseConfiguration(metadata)



# A very simple setup script to create 2 executables.
#
# hello.py is a simple "hello, world" type program, which alse allows
# to explore the environment in which the script runs.
#
# test_wx.py is a simple wxPython program, it will be converted into a
# console-less program.
#
# If you don't have wxPython installed, you should comment out the
#   windows = ["test_wx.py"]
# line below.
#
#
# Run the build process by entering 'setup.py py2exe' or
# 'python setup.py py2exe' in a console prompt.
#
# If everything works well, you should find a subdirectory named 'dist'
# containing some files, among them hello.exe and test_wx.exe.


from distutils.core import setup
import py2exe

# The manifest enables the use of the XP theme in wxWidgets.
# How logical and intuitive.
manifest = """
<?xml version="1.0" encoding="UTF-8" standalone="yes"?>
<assembly xmlns="urn:schemas-microsoft-com:asm.v1" manifestVersion="1.0">
<assemblyIdentity
    version="0.64.1.0"
    processorArchitecture="x86"
    name="Controls"
    type="win32"
/>
<description>Snowberry</description>
<dependency>
    <dependentAssembly>
        <assemblyIdentity
            type="win32"
            name="Microsoft.Windows.Common-Controls"
            version="6.0.0.0"
            processorArchitecture="X86"
            publicKeyToken="6595b64144ccf1df"
            language="*"
        />
    </dependentAssembly>
</dependency>
</assembly>
"""

opts = {
    "py2exe": {
        "includes": "webbrowser",
        'bundle_files': 1,
        'compressed': 1,
        'optimize': 2
    }
}

"""
installs manifest and icon into the .exe
but icon is still needed as we open it
for the window icon (not just the .exe)
changelog and logo are included in dist
"""

setup(
    version = "1.2.0",
    description = "Doomsday Engine Launcher",
    name = "Snowberry",
    options = opts,

    windows = [
        {
            "script": "snowberry.py",
            "icon_resources": [(1, "graphics/snowberry.ico")],
            "other_resources": [(24, 1, manifest)]
        }
    ],
    data_files= [ 
        ( '.', ['graphics/snowberry.ico'] ) 
    ]
)



# -*- coding: iso-8859-1 -*-
# $Id$
# Snowberry: Extensible Launcher for the Doomsday Engine
#
# Copyright (C) 2004, 2005
#   Jaakko Keränen <jaakko.keranen@iki.fi>
#   Antti Kopponen <antti.kopponen@tut.fi>
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not: http://www.opensource.org/

## @file profilelist.py Profile List

import ui, events, widgets
import sb.util.dialog
import sb.widget.button as wg
import sb.widget.text as wt
import sb.widget.list
import sb.profdb as pr
import sb.confdb as st
import language
import paths
from ui import ALIGN_HORIZONTAL

USE_MINIMAL_MODE = st.getSystemBoolean('profile-minimal-mode')

profileList = None
deleteButton = None
dupeButton = None

# If set to true, the profile list won't be updated on notifications.
profileListDisabled = False

# Popup menus.
defaultsMenu = ['reset-profile', '-', 'new-profile', 'unhide-profiles']

normalMenu = ['hide-profile',
              'rename-profile',
              'duplicate-profile',
              'reset-profile',
              'delete-profile',
              '-',
              'new-profile',
              'unhide-profiles']

systemMenu = ['hide-profile',
              'rename-profile',
              'duplicate-profile',
              'reset-profile',
              '-',
              'new-profile',
              'unhide-profiles']


def getIconSize():
    """Determines the size of profile icons.
    @return Size of profile icons in pixels, as a string."""
    if st.getSystemBoolean('profile-large-icons'):
        return 50
    else:
        return 27


def makeHTML(name, iconName, gameName=None, boldName=False):
    # The icon for the game component.
    iconSize = getIconSize()
    iconPath = widgets.iconManager.getBitmapPath(iconName, (iconSize, iconSize))
    
    if boldName:
        nameStyle = ('<b>', '</b>')
    else:
        nameStyle = ('', '')

    if st.getSystemBoolean('profile-large-icons'): 
        padding = 5
    else:
        padding = 3
    
    html = '<table width="100%%" border=0 cellspacing=2 cellpadding=%i>' % padding
    html += '<tr><td width="%i"' % iconSize + '><img width="%i"' % iconSize
    html += ' height="%i"' % iconSize + ' src="' + iconPath + '"></td><td>' 
    html += '<font size="+1">' + nameStyle[0] + name + nameStyle[1] + '</font>'
    if gameName:
        html += '<br><font color="#808080" size="-1">' + gameName + '</font>' 
    html += '</td></tr></table>'
    return html 


def makeProfileHTML(profile):
    if profile is pr.getDefaults():
        return makeHTML(profile.getName(), 'defaults', boldName=True)
        #iconPath = paths.findBitmap('defaults')
        #iconSize = iconSizeString()
        #return '<table width="100%" border=0 cellspacing=2 cellpadding=5>' + \
        #   '<tr><td width="' + iconSize + \
        #   '"><img width="%s" height="%s"' % (iconSize, iconSize) + \
        #   ' src="' + iconPath + \
        #   '"><td align=left><font size="+1"><b>' + profile.getName() + \
        #   '</b></font></table>'
    else:
        game = 'game-undefined'
        for c in profile.getComponents():
            if c[:5] == 'game-':
                game = c
                break

        return makeHTML(profile.getName(), language.translate(game + '-icon'),
                        language.translate(game))


def init():
    # Create the profile list and the profile control buttons.
    area = ui.getArea(ui.PROFILES)

    global profileList
    
    area.setBorder(0)
    
    if USE_MINIMAL_MODE:
        area.setWeight(1)
        area.addSpacer()
        area.setWeight(3)
        profileList = area.createDropList('profile-list')
        profileList.setSorted()
        area.setWeight(1)
        area.addSpacer()
        
    else:
        # Normal profile list mode.
        area.setWeight(1)
        profileList = area.createFormattedList("profile-list")

        if not st.getSystemBoolean('profile-hide-buttons'):
            # This should be a small button.
            area.setWeight(0)
            area.setBorder(3)
            controls = area.createArea(alignment=ALIGN_HORIZONTAL, border=2)
            controls.setExpanding(False)
            #area.setExpanding(False)
            controls.setWeight(0)
            controls.createButton('new-profile', wg.Button.STYLE_MINI)

            global deleteButton
            deleteButton = controls.createButton('delete-profile',
                                                 wg.Button.STYLE_MINI)

            global dupeButton
            dupeButton = controls.createButton('duplicate-profile',
                                               wg.Button.STYLE_MINI)

    # Set the title graphics.
    global bannerImage
    try:
        area = ui.getArea(ui.TITLE)
        bannerImage = area.createImage('banner-default')
    except:
        # There is no title area.
        bannerImage = None

    # Register a listener for notifications.
    events.addNotifyListener(notifyHandler, ['quit', 'language-changed',
                                             'profile-updated',
                                             'active-profile-changed',
                                             'profile-list-selected'])
                                             
    # Register a listener for commands.
    events.addCommandListener(commandHandler, ['freeze', 'unfreeze',
                                               'new-profile', 
                                               'rename-profile',
                                               'reset-profile', 
                                               'delete-profile',
                                               'duplicate-profile',
                                               'hide-profile', 
                                               'unhide-profiles'])

    # Commands for the menu.
    ui.addMenuCommand(ui.MENU_PROFILE, 'new-profile', group=ui.MENU_GROUP_PROFDB)
    ui.addMenuCommand(ui.MENU_PROFILE, 'unhide-profiles', group=ui.MENU_GROUP_PROFDB)

    ui.addMenuCommand(ui.MENU_PROFILE, 'reset-profile', group=ui.MENU_GROUP_PROFILE)
    ui.addMenuCommand(ui.MENU_PROFILE, 'rename-profile', group=ui.MENU_GROUP_PROFILE)
    ui.addMenuCommand(ui.MENU_PROFILE, 'duplicate-profile', group=ui.MENU_GROUP_PROFILE)
    ui.addMenuCommand(ui.MENU_PROFILE, 'hide-profile', group=ui.MENU_GROUP_PROFILE)
    ui.addMenuCommand(ui.MENU_PROFILE, 'delete-profile', group=ui.MENU_GROUP_PROFILE)


def addListItem(profile, toIndex=None):
    if not USE_MINIMAL_MODE:
        profileList.addItem(profile.getId(), makeProfileHTML(profile), toIndex)
    else:
        profileList.addItem(profile.getId(), toIndex)


def notifyHandler(event):
    "This is called when a Notify event is broadcasted."

    # We are interested in notifications that concern profiles.
    if event.hasId('quit'):
        # Save the current profile configuration.
        pr.save()

    elif event.hasId('language-changed'):
        # Just update the Defaults, because it's the only profile whose name
        # is translated.
        addListItem(pr.getDefaults())

    elif event.hasId('profile-updated'):
        # A profile has been loaded or updated.  Make sure it's in the
        # list.
        p = event.getProfile()

        # The Defaults profile should be first in the list.
        if p is pr.getDefaults():
            destIndex = 0
        else:
            # The new profile will be added to wherever the widget
            # wants to put it.
            destIndex = None

        if profileList.hasItem(p.getId()):
            # This already exists in the list.
            if p.isHidden():
                # Remove it completely.
                profileList.removeItem(p.getId())

                # Update the selected item since the active one is now
                # hidden.
                pr.setActive(pr.get(profileList.getSelectedItem()))
            else:
                # Just update the item.
                addListItem(p)

        elif not p.isHidden():
            # The item does not yet exist in the list.
            addListItem(p, toIndex=destIndex)

    elif event.hasId('active-profile-changed') and not profileListDisabled:
        # Highlight the correct profile in the list.
        profileList.selectItem(pr.getActive().getId())

        if pr.getActive() is pr.getDefaults():
            if deleteButton: deleteButton.disable()
            if dupeButton: dupeButton.disable()
            ui.disableMenuCommand('rename-profile')
            ui.disableMenuCommand('delete-profile')
            ui.disableMenuCommand('hide-profile')
            ui.disableMenuCommand('duplicate-profile')
            profileList.setPopupMenu(defaultsMenu)
        else:
            isSystem = pr.getActive().isSystemProfile()
            if deleteButton: deleteButton.enable(not isSystem)
            if dupeButton: dupeButton.enable()
            ui.enableMenuCommand('rename-profile')
            ui.enableMenuCommand('delete-profile', not isSystem)
            ui.enableMenuCommand('hide-profile')
            ui.enableMenuCommand('duplicate-profile')
            if isSystem:
                menu = systemMenu
            else:
                menu = normalMenu
            profileList.setPopupMenu(menu)

        # Update the banner image.
        if bannerImage:
            bannerImage.setImage(pr.getActive().getBanner())

    elif event.hasId('profile-list-selected'):
        # Change the currently active profile.
        p = pr.get(event.getSelection())
        pr.setActive(p)

        # Double-clicking causes a Play command.
        if event.isDoubleClick() and p is not pr.getDefaults():
            events.sendAfter(events.Command('play'))


def commandHandler(event):
    """This is called when a Command event is broadcasted."""

    global profileListDisabled

    if event.hasId('freeze'):
        profileListDisabled = True

    elif event.hasId('unfreeze'):
        profileListDisabled = False
        #notifyHandler(events.Notify('active-profile-changed'))
        pr.refresh()

    elif event.hasId('new-profile'):
        dialog, area = sb.util.dialog.createButtonDialog(
            'new-profile-dialog',
            ['cancel', 'ok'], 'ok', resizable=False)
        dialog.disableWidget('ok')

        entry = area.createArea(alignment=ALIGN_HORIZONTAL)
        entry.setExpanding(False)

        # The name of the profile.
        entry.setWeight(1)
        entry.setBorder(4, ui.BORDER_RIGHT)
        entry.createText('new-profile-name', ':', align=wt.Text.RIGHT)
        entry.setBorder(0)
        entry.setWeight(2)
        nameField = entry.createTextField('')
        nameField.focus()

        # Only enable the OK button if there is something in the name field.
        def buttonEnabler():
            dialog.enableWidget('ok', len(nameField.getText()) > 0)

        nameField.addReaction(buttonEnabler)

        # The game component.
        entry = area.createArea(alignment=ALIGN_HORIZONTAL)
        entry.setExpanding(False)

        entry.setWeight(1)
        entry.setBorder(4, ui.BORDER_RIGHT)
        entry.createText('new-profile-game', ':', align=wt.Text.RIGHT)
        entry.setBorder(0)
        entry.setWeight(2)
        gameDrop = entry.createDropList('')

        for game in st.getGameComponents():
            gameDrop.addItem(game.getId())

        gameDrop.sortItems()

        # Select the first one.
        gameDrop.selectItem(gameDrop.getItems()[0])

        if dialog.run() == 'ok':
            # Get the values from the controls.
            pr.create(nameField.getText(), gameDrop.getSelectedItem())

    elif event.hasId('rename-profile'):
        dialog, area = sb.util.dialog.createButtonDialog(
            'rename-profile-dialog',
            ['cancel', 'ok'], 'ok', resizable=False)

        prof = pr.getActive()

        # A text field of the new name.
        entry = area.createArea(alignment=ALIGN_HORIZONTAL)
        entry.setExpanding(False)

        # The name of the profile.
        entry.setWeight(1)
        entry.setBorder(4, ui.BORDER_RIGHT)
        entry.createText('rename-profile-name', ':', align=wt.Text.RIGHT)
        entry.setBorder(0)
        entry.setWeight(2)
        nameField = entry.createTextField('')
        nameField.setText(prof.getName())
        nameField.focus()

        # Only enable the OK button if there is something in the name
        # field.
        def buttonEnabler():
            dialog.enableWidget('ok', len(nameField.getText()) > 0)

        nameField.addReaction(buttonEnabler)

        if dialog.run() == 'ok':
            prof.setName(nameField.getText())
            # The profile list needs to be updated, too.
            events.send(events.ProfileNotify(prof))
            pr.refresh()

    elif event.hasId('reset-profile'):
        dialog, area = sb.util.dialog.createButtonDialog(
            'reset-profile-dialog',
            ['cancel', 'reset-profile-values', 'reset-profile-addons', 
            'reset-profile-everything'], 
            'cancel', resizable=False)

        text = language.translate('reset-profile-query')
        message = area.createRichText(
            language.expand(text, pr.getActive().getName()))

        # Accept these as dialog-closing commands.
        dialog.addEndCommand('reset-profile-values')
        dialog.addEndCommand('reset-profile-addons')
        dialog.addEndCommand('reset-profile-everything')
        
        result = dialog.run()
        if result == 'cancel':
            return

        resetValues = True
        resetAddons = True
        if result == 'reset-profile-values':
            resetAddons = False
        elif result == 'reset-profile-addons':
            resetValues = False
        pr.reset(pr.getActive().getId(), resetValues, resetAddons)
        pr.refresh()

    elif event.hasId('delete-profile'):
        # If this is a system profile, just hide it.
        if pr.getActive().isSystemProfile():
            pr.hide(pr.getActive().getId())
            return
        
        dialog, area = sb.util.dialog.createButtonDialog(
            'delete-profile-dialog',
            ['no', 'yes'], 'no', resizable=False)

        text = language.translate('delete-profile-query')
        area.createRichText(language.expand(text, pr.getActive().getName()))

        if dialog.run() == 'yes':
            # Get the values from the controls.
            pr.remove(pr.getActive().getId())

    elif event.hasId('duplicate-profile'):
        dialog, area = sb.util.dialog.createButtonDialog(
            'duplicate-profile-dialog',
            ['cancel', 'ok'], 'ok', resizable=False)

        text = language.translate('duplicating-profile')
        area.setWeight(1)
        area.createRichText(language.expand(text, pr.getActive().getName()))

        area.setWeight(1)
        entry = area.createArea(alignment=ALIGN_HORIZONTAL)
        entry.setExpanding(False)

        # The name of the profile.
        entry.setWeight(1)
        entry.setBorder(4, ui.BORDER_RIGHT)
        entry.createText('new-profile-name', ':', align=wt.Text.RIGHT)
        entry.setBorder(0)
        entry.setWeight(3)
        nameField = entry.createTextField('')
        nameField.setText(pr.getActive().getName())
        nameField.focus()

        # Only enable the OK button if there is something in the name field.
        def buttonEnabler():
            dialog.enableWidget('ok', len(nameField.getText()) > 0)

        nameField.addReaction(buttonEnabler)

        if dialog.run() == 'ok':
            pr.duplicate(pr.getActive().getId(), nameField.getText())

    elif event.hasId('hide-profile'):
        pr.hide(pr.getActive().getId())

    elif event.hasId('unhide-profiles'):
        # Display a dialog bog for unhiding hidden profiles.
        hiddenProfiles = pr.getProfiles(lambda p: p.isHidden())

        dialog, area = sb.util.dialog.createButtonDialog(
            'unhide-profile-dialog',
            ['cancel', 'ok'], 'ok', resizable=False)

        area.setWeight(0)
        area.createText('unhiding-profiles')

        area.setWeight(3)
        area.setBorder(0)
        profList = area.createList('', sb.widget.list.List.STYLE_CHECKBOX)
        profList.setMinSize(50, 150)

        def selectAll():
            # Check the entire list.
            for item in profList.getItems():
                profList.checkItem(item, True)

        def clearAll():
            # Uncheck the entire list.
            for item in profList.getItems():
                profList.checkItem(item, False)

        area.setWeight(0)
        controls = area.createArea(alignment=ALIGN_HORIZONTAL,
                                   border=2)
        controls.setWeight(0)
        button = controls.createButton('unhide-select-all', 
            style=sb.widget.button.Button.STYLE_MINI)
        button.addReaction(selectAll)
        button.resizeToBestSize()
        button = controls.createButton('unhide-clear-all', 
            style=sb.widget.button.Button.STYLE_MINI)
        button.addReaction(clearAll)
        button.resizeToBestSize()

        for prof in hiddenProfiles:
            profList.addItem(prof.getId())

        if dialog.run() == 'ok':
            # Unhide all the selected items.
            selected = profList.getSelectedItems()
            for sel in selected:
                pr.show(sel)

            if len(selected):
                # Select the first one that was shown.
                pr.setActive(pr.get(selected[0]))



# -*- coding: iso-8859-1 -*-
# $Id: main.py 4303 2007-01-01 19:06:40Z skyjake $
# Snowberry: Extensible Launcher for the Doomsday Engine
#
# Copyright (C) 2004, 2005
#   Jaakko Keränen <jaakko.keranen@iki.fi>
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not: http://www.opensource.org/

## @file inspector.py Addon Inspector Dialog

import os, time, string
import paths, events, ui, language
import sb.util.dialog
import sb.widget.button as wg
import sb.widget.list
import sb.profdb as pr
import sb.confdb as st
import sb.aodb as ao
import sb.addon


def run(addon):
    """Show a dialog box that contains a lot of information (all there
    is to know) about the addon.

    @param An addons.Addon object.
    """
    ident = addon.getId()
    
    dialog, area = sb.util.dialog.createButtonDialog(
        'addon-inspector-dialog',
         ['ok'], 'ok', size=(570, 450))

    msg = ""

    msg += '<h3>' + language.translate(ident) + '</h3>'
    if language.isDefined(ident + '-readme'):
        msg += "<p>" + language.translate(ident + '-readme')

    def makeField(header, content):
        return '<tr><td width="20%" bgcolor="#E8E8E8" align="right"><b>' + header + \
               '</b></td><td width="80%">' + content + '</td></tr>'

    beginTable = '<p><table width="100%" border=0 cellpadding=6 cellspacing=0>'
    endTable = '</table>'

    #
    # General Information
    #
    msg += beginTable
    msg += makeField('Summary:', language.translate(ident + '-summary', '-'))
    msg += makeField('Version:', language.translate(ident + '-version', '-'))
    msg += makeField('Last Modified:',
                     time.strftime("%a, %d %b %Y %H:%M:%S",
                                   time.localtime(addon.getLastModified())))
    msg += makeField('Author(s):', language.translate(ident + '-author', '-'))
    msg += makeField('Contact:', language.translate(ident + '-contact', '-'))
    msg += makeField('Copyright:',
                     language.translate(ident + '-copyright', '-'))
    msg += makeField('License:', language.translate(ident + '-license', '-'))
    msg += makeField('Category:', addon.getCategory().getPath())
    msg += makeField('Format:', language.translate(addon.getType()))
    msg += makeField('Identifier:', addon.getId())
    msg += endTable

    #
    # Raw Dependencies
    #
    msg += '<h3>Dependencies</h3>' + beginTable
    allDeps = []
    
    # Excluded categories.
    dep = []
    for category in addon.getExcludedCategories():
        dep.append(category.getPath())
    allDeps.append(('Excluded Categories:', dep))

    # Keywords.
    for type, label in [(sb.addon.EXCLUDES, 'Excludes:'),
                        (sb.addon.REQUIRES, 'Requires:'),
                        (sb.addon.PROVIDES, 'Provides:'),
                        (sb.addon.OFFERS, 'Offers:')]:
        # Make a copy of the list so we can modify it.
        dep = []
        for kw in addon.getKeywords(type):
            dep.append(kw)
        allDeps.append((label, dep))

    # Create a table out of each dependency type.
    for heading, listing in allDeps:
        msg += makeField(heading, string.join(listing, '<br>'))

    msg += endTable

    #
    # Content Analysis
    # 
    msg += '<h3>Contents</h3>' + beginTable
    msg += makeField('Content Path:', addon.getContentPath())
    msg += endTable
    
    #msg += "<p>format-specific data"
    #msg += "<br>content analysis, size"
    #msg += "<br>list of files, if a bundle"

    #msg += '<h3>Identifiers</h3>'
    #msg += "<br>all internal identifiers used by the addon"

    #msg += '<h3>Metadata</h3>'
    #msg += "<br>metadata analysis, source"

    text = area.createFormattedText()
    text.setMinSize(500, 200)
    text.setText(msg)
    dialog.run()        





# -*- coding: iso-8859-1 -*-
# $Id$
# Snowberry: Extensible Launcher for the Doomsday Engine
#
# Copyright (C) 2004, 2005
#   Jaakko Keränen <jaakko.keranen@iki.fi>
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not: http://www.opensource.org/

## @file logger.py Warning/Error Logger
##
## This module collections warnings and errors and displays them when
## asked to.

import sys, traceback
import language, string


# An array with the currently collection issues.
allIssues = []

# Levels of severity.
LOW = 0
MEDIUM = 1
HIGH = 2


def add(severity, identifier, *args):
    """Add a new issue to the list of issues waiting to be displayed.

    @param severity The severity of the issue.  One of the following
    constants:
    - LOW 
    - MEDIUM
    - HIGH

    @param identifier Identifier of the issue.

    @param args An array of arguments.  These will be used to expand
    placeholders in the translated version of the identifier.
    """
    global allIssues

    allIssues.append((severity, identifier, list(args)))
    
    
def immediate(severity, identifier, *args):
    """Show a logger issue immediately. See logger.add() for the params."""
    
    add(severity, identifier, *args)
    show()


def show():
    """Show all the issues in a dialog.  The list of issues will be
    automatically cleared.  If there are no issues to report, nothing
    is done."""

    global allIssues

    if len(allIssues) == 0:
        # Nothing to report.
        return

    import sb.util.dialog
    dialog, area = sb.util.dialog.createButtonDialog(
        'issue-dialog',
        ['ok'], 'ok')

    message = area.createFormattedText()
    message.setMinSize(500, 300)

    msg = ''

    bgColor = ['#90E090', '#FFFF80', '#E05050']
    fgColor = ['black', 'black', 'white']
    severeName = ['Note', 'Warning', 'Error']

    isFirst = True

    # Compose a HTML formatted version of each issue.
    for severe in [HIGH, MEDIUM, LOW]:
        # Show the important issues first.
        for issue in allIssues:
            if issue[0] != severe:
                continue

            if not isFirst:
                msg += '<hr>'
            isFirst = False

            msg += '<table width="100%" border=0 cellspacing=3 cellpadding=4>'
            msg += '<tr><td bgcolor="%s" width="20%%" align=center>' % \
                   bgColor[issue[0]]
            msg += '<font color="%s"><b>' % fgColor[issue[0]] + \
                   severeName[issue[0]] + '</b></font>'
            msg += '<td width="80%"><h3>' + language.expand(
                language.translate(issue[1]), *issue[2])
            msg += '</h3>' + language.expand(
                language.translate(issue[1] + '-text'), *issue[2])
            msg += '</table>'

    message.setText(msg)
    dialog.run()
        
    # Only show each issue once.
    allIssues = []


def formatTraceback():
    """Make a rough HTML formatting of the system traceback.   
    @return Traceback as a string of HTML text."""    
    
    str = string.join(traceback.format_exception_only(sys.exc_type, sys.exc_value)) + '\n'
    
    str += string.join(traceback.format_list(
                       traceback.extract_tb(sys.exc_traceback)), '<br>')
    str = str.replace('\n', '<br>')
    str = str.replace(' ', '&nbsp;')
    return '<font size="-3"><tt>' + str + '</tt></font>'



import os, glob, shutil, time
import build_number
import config
import utils

def log_filename(package, osIdent, ext='txt.gz'):
    return 'buildlog-%s-%s.%s' % (package, osIdent, ext)


class Event:
    """Build event. Manages the contents of a single build directory under
    the event directory."""
    
    def __init__(self, build=None, latestAvailable=False):
        """Any .txt logs present in the build directory are compressed into
        a combined .txt.gz (one per package)."""

        if latestAvailable:
            # Look for the latest build.
            build = int(build_number.todays_build())
            while not os.path.exists(os.path.join(config.EVENT_DIR, 'build%i' % build)):
                build -= 1
                if build == 0: raise Exception("No builds available")
        
        if build is None:
            # Use today's build number.
            self.name = 'build' + build_number.todays_build()
            self.num = int(self.name[5:])
        elif type(build) == int:
            self.name = 'build' + str(build)
            self.num = build
        elif type(build) == str:
            if build[:5] != 'build': 
                raise Exception("Event build name must begin with 'build'")
            self.name = build
            self.num = int(build[5:])

        # Where the build is located.
        self.buildDir = os.path.join(config.EVENT_DIR, self.name)
        
        self.packages = ['doomsday', 'doomsday_app', 'doomsday_shell_app', 'fmod']
        
        self.packageName = {'doomsday':           'Doomsday',
                            'doomsday_apps':      'OS X Apps',
                            'doomsday_app':       'Doomsday Engine.app',
                            'doomsday_shell_app': 'Doomsday Shell.app',
                            'fmod':               'FMOD Ex Audio Plugin'}
        
        if self.num >= 816: # Added Mac OS X 10.8.
            # Platforms:  Name                              File ext          sys_id()
            self.oses = [('Windows (x86)',                 '.exe',           'win32-32bit'),
                         ('OS X 10.8+ (x86_64)',           ('.dmg', 'macx8.dmg'), 'macx8-64bit'),
                         ('OS X 10.6+ (x86_64/i386)',      ('mac10_6.dmg', 'macx6.dmg'), 'darwin-64bit'),
                         ('OS X 10.4+ (ppc/i386)',         '32bit.dmg',      'darwin-32bit'),
                         ('Ubuntu (x86_64)',               'amd64.deb',      'linux2-64bit'),
                         ('Ubuntu (x86)',                  'i386.deb',       'linux2-32bit'),
                         ('Source',                        '.tar.gz',        'source')]

            # Remove obsolete OS X versions:
            if self.has_version():
                if utils.version_cmp(self.version_base(), '1.11') >= 0:
                    del self.oses[3] # no more OS X 10.4
                if self.num >= 1212 and utils.version_cmp(self.version_base(), '1.15') >= 0:
                    del self.oses[2] # no more OS X 10.6
                
        elif self.num >= 778: # Mac distribution naming was changed.
            # Platforms:  Name                            File ext     sys_id()
            self.oses = [('Windows (x86)',                '.exe',      'win32-32bit'),
                         ('Mac OS X 10.6+ (x86_64/i386)', '.dmg',      'darwin-64bit'),
                         ('Mac OS X 10.4+ (ppc/i386)',    '32bit.dmg', 'darwin-32bit'),
                         ('Ubuntu (x86_64)',              'amd64.deb', 'linux2-64bit'),
                         ('Ubuntu (x86)',                 'i386.deb',  'linux2-32bit')]

        else:
            # Platforms:  Name                            File ext     sys_id()
            self.oses = [('Windows (x86)',                '.exe',      'win32-32bit'),
                         ('Mac OS X 10.4+ (ppc/i386)',    '.dmg',      'darwin-32bit'),
                         ('Mac OS X 10.6+ (x86_64/i386)', '64bit.dmg', 'darwin-64bit'),
                         ('Ubuntu (x86)',                 'i386.deb',  'linux2-32bit'),
                         ('Ubuntu (x86_64)',              'amd64.deb', 'linux2-64bit')]            

        self.platId = {'win32-32bit':  'win-x86',
                       'darwin-32bit': 'mac10_4-x86-ppc',
                       'darwin-64bit': 'mac10_6-x86-x86_64',
                       'macx8-64bit':  'mac10_8-x86_64',
                       'linux2-32bit': 'linux-x86',
                       'linux2-64bit': 'linux-x86_64',
                       'source':       'source'}

        # Prepare compiler logs present in the build dir.
        self.compress_logs()
        
    def package_type(self, name):
        pkg = self.package_from_filename(name)
        if pkg == 'fmod':
            return 'plugin'
        else:
            return 'distribution'

    def package_from_filename(self, name):
        if 'apps-macx' in name: return 'doomsday_apps'
        if name.endswith('.zip'):
            if 'doomsday_osx' in name:
                return 'doomsday_app'
            if 'doomsday_shell_osx' in name:
                return 'doomsday_shell_app'
        if 'fmod' in name:
            return 'fmod'
        else:
            return 'doomsday'        
    
    def os_from_filename(self, name):
        found = None        
        for n, osExt, ident in self.oses:
            if type(osExt) == 'tuple':
                exts = osExt
            else:
                exts = [osExt]
            for ext in exts:
                if name.endswith(ext) or ident in name:
                    found = (n, ext, ident)
            if n.startswith('OS X 10.') and name.endswith('.zip'):
                osx = '_osx' + n[8] + '_'
                if osx in name:
                    found = (n, ext, ident)
        return found
                
    def version_from_filename(self, name):
        ver = self.extract_version_from_filename(name)        
        if not ver and self.package_from_filename(name) != 'fmod':
            # Fall back to the event version, if it exists.
            ev = self.version()
            if ev: return ev
        return ver

    def extract_version_from_filename(self, name):
        if '_osx' in name:
            pos = name.find('_osx') + 4
        else:
            pos = 0
        pos = name.find('_', pos)
        if pos < 0: return None
        dash = name.find('-', pos + 1)
        us = name.find('_', pos + 1)
        if dash > 0 and us < 0:
            return name[pos+1:dash]
        elif us > 0 and dash < 0:
            return name[pos+1:us]
        elif dash > 0 and us > 0:
            return name[pos+1:min(dash, us)]
        return name[pos+1:name.rfind('.', pos)]
        
    def tag(self):
        return self.name
        
    def version_base(self):
        ver = self.version()
        if '-' in ver: ver = ver[:ver.find('-')]
        return ver
        
    def version(self):
        fn = self.file_path('version.txt')
        if os.path.exists(fn): return file(fn).read().strip()
        return None
        
    def has_version(self):
        return os.path.exists(self.file_path('version.txt'))
        
    def name(self):
        return self.name
        
    def number(self):
        """Returns the event's build number as an integer."""
        return self.num
        
    def path(self):
        return self.buildDir
        
    def file_path(self, fileName):
        return os.path.join(self.buildDir, fileName)
        
    def clean(self):
        # Make sure we have a clean directory for this build.
        if os.path.exists(self.buildDir):
            # Kill it and recreate.
            shutil.rmtree(self.buildDir, True)
        os.mkdir(self.buildDir)        
        
    def list_package_files(self):
        files = glob.glob(os.path.join(self.buildDir, '*.dmg')) + \
                glob.glob(os.path.join(self.buildDir, '*.exe')) + \
                glob.glob(os.path.join(self.buildDir, '*.deb')) + \
                glob.glob(os.path.join(self.buildDir, '*.tar.gz'))
            
        if self.num > 1201:
            # Zipped apps added.
            files += glob.glob(os.path.join(self.buildDir, '*.zip'))

        return [os.path.basename(f) for f in files]

    def timestamp(self):
        """Looks through the files of the build and returns the timestamp
        for the oldest file."""
        oldest = os.stat(self.buildDir).st_ctime

        for fn in os.listdir(self.buildDir):
            t = os.stat(os.path.join(self.buildDir, fn))
            if int(t.st_mtime) < oldest:
                oldest = int(t.st_mtime)

        return oldest        
        
    def text_timestamp(self):
        return time.strftime(config.RFC_TIME, time.gmtime(self.timestamp()))
    
    def text_summary(self):
        """Composes a textual summary of the event."""
        
        msg = "The build event was started on %s." % self.text_timestamp()

        msg += ' It'

        pkgCount = len(self.list_package_files())

        changesName = self.file_path('changes.html')
        commitCount = 0
        if os.path.exists(changesName):
            commitCount = utils.count_word('<li>', file(changesName).read())
        if commitCount:
            moreThan = ''
            if commitCount == 100: moreThan = 'more than '
            msg += " contains %s%i commits and" % (moreThan, commitCount)

        msg += " produced %i package%s." % \
            (pkgCount, 's' if (pkgCount != 1) else '')

        return msg
        
    def compress_logs(self):
        if not os.path.exists(self.buildDir): return
        
        """Combines the stdout and stderr logs for a package and compresses
        them with gzip (requires gzip on the system path)."""
        for package in self.packages:
            for osName, osExt, osIdent in self.oses:
                names = glob.glob(self.file_path('%s-*-%s.txt' % (package, osIdent)))
                if not names: continue
                # Join the logs into a single file.
                combinedName = self.file_path('buildlog-%s-%s.txt' % (package, osIdent))
                combined = file(combinedName, 'wt')
                for n in sorted(names):
                    combined.write(file(n).read() + "\n\n")
                    # Remove the original log.
                    os.remove(n)
                combined.close()            
                os.system('gzip -f9 %s' % combinedName)        
                
    def download_uri(self, fn):
        # Available on SourceForge?
        if self.number() >= 350 and (fn.endswith('.exe') or fn.endswith('.deb') or fn.endswith('.dmg') or fn.endswith('.zip')):
            if self.release_type() == 'stable':
                return "http://sourceforge.net/projects/deng/files/Doomsday%%20Engine/%s/%s/download" \
                    % (self.version_base(), fn)
            else:
                return "http://sourceforge.net/projects/deng/files/Doomsday%%20Engine/Builds/%s/download" % fn
        # Default to the old location.
        return "%s/%s/%s" % (config.BUILD_URI, self.name, fn)
        
    def download_fallback_uri(self, fn):
        return "%s/%s/%s" % (config.BUILD_URI, self.name, fn)
                
    def compressed_log_filename(self, binaryFn):
        return log_filename(self.package_from_filename(binaryFn), 
                            self.os_from_filename(binaryFn)[2])

    def sort_by_package(self, binaries):
        """Returns the list of binaries sorted by package."""
        pl = []
        for bin in binaries:
            pl.append((self.package_from_filename(bin), bin))
        pl.sort()
        return [bin for pkg, bin in pl]
               
    def html_table_log_issues(self, logName):
        # Link to the compressed log.
        msg = '<td><a href="%s">txt.gz</a>' % self.download_uri(logName)

        # Show a traffic light indicator based on warning and error counts.              
        errors, warnings = utils.count_log_issues(self.file_path(logName))
        form = '<td bgcolor="%s" style="text-align:center;">'
        if errors > 0:
            msg += form % '#ff4444' # red
        elif warnings > 0:
            msg += form % '#ffee00' # yellow
        else:
            msg += form % '#00ee00' # green
        msg += str(errors + warnings)        

        return msg
                
    def html_description(self, encoded=True):
        """Composes an HTML build report."""

        name     = self.name
        buildDir = self.buildDir
        oses     = self.oses

        msg = '<p>' + self.text_summary() + '</p>'

        # What do we have here?
        files = self.list_package_files()

        # Print out the matrix.
        msg += '<p><table cellspacing="4" border="0">'
        msg += '<tr style="text-align:left;"><th>OS<th>Binary<th>Logs<th>Issues</tr>'

        for osName, osExt, osIdent in oses:
            isFirst = True
            # Find the binaries for this OS.
            binaries = []
            for f in files:
                if self.os_from_filename(f)[2] == osIdent:
                    binaries.append(f)

            if not binaries:
                # Nothing available for this OS.
                msg += '<tr><td>' + osName + '<td>n/a'
                
                # Do we have a log?
                logName = log_filename(self.packages[0], osIdent)
                if os.path.exists(self.file_path(logName)):
                    msg += self.html_table_log_issues(logName)
                
                msg += '</tr>'
                continue

            # List all the binaries. One row per binary.
            for binary in self.sort_by_package(binaries):
                msg += '<tr><td>'
                if isFirst:
                    msg += osName
                    isFirst = False
                msg += '<td>'
                msg += '<a href="%s">%s</a>' % (self.download_uri(binary), binary)

                # Status of the log.
                logName = self.compressed_log_filename(binary)
                if not os.path.exists(self.file_path(logName)):
                    msg += '</tr>'
                    continue                            

                # Link to the compressed log.
                msg += self.html_table_log_issues(logName)

            msg += '</tr>'

        msg += '</table></p>'

        # Changes.
        chgFn = self.file_path('changes.html')
        if os.path.exists(chgFn):
            if utils.count_word('<li>', file(chgFn).read()):
                msg += '<h2>Commits</h2>' + file(chgFn, 'rt').read()

        # Enclose it in a CDATA block if needed.
        if encoded: return '<![CDATA[' + msg + ']]>'    
        return msg
        
    def release_type(self):
        """Returns the release type as a lower-case string."""
        fn = self.file_path('releaseType.txt')
        if os.path.exists(fn):
            return file(fn).read().lower().strip()
        return 'unstable' # Default assumption.
        
    def xml_log(self, logName):
        msg = '<compileLogUri>%s</compileLogUri>' % self.download_uri(logName)
        errors, warnings = utils.count_log_issues(self.file_path(logName))
        msg += '<compileWarnCount>%i</compileWarnCount>' % warnings
        msg += '<compileErrorCount>%i</compileErrorCount>' % errors
        return msg
    
    def release_notes_uri(self, version):
        return "http://dengine.net/dew/index.php?title=Doomsday_version_" + version
        
    def changelog_uri(self, version):
        if self.release_type() == 'stable':
            return self.release_notes_uri(version)
        else:
            return "http://dengine.net/" + self.name          

    def xml_description(self):
        msg = '<build>'
        msg += '<uniqueId>%i</uniqueId>' % self.number()
        msg += '<startDate>%s</startDate>' % self.text_timestamp()
        msg += '<authorName>%s</authorName>' % config.BUILD_AUTHOR_NAME
        msg += '<authorEmail>%s</authorEmail>' % config.BUILD_AUTHOR_EMAIL
        msg += '<releaseType>%s</releaseType>' % self.release_type()
        files = self.list_package_files()
        msg += '<packageCount>%i</packageCount>' % len(files)
        
        # These logs were already linked to.
        includedLogs = []
        
        distribVersion = None
        
        # Packages.
        for fn in files:
            msg += '<package type="%s">' % self.package_type(fn)
            msg += '<name>%s</name>' % self.packageName[self.package_from_filename(fn)]
            msg += '<version>%s</version>' % self.version_from_filename(fn)
            msg += '<platform>%s</platform>' % self.platId[self.os_from_filename(fn)[2]]
            msg += '<downloadUri>%s</downloadUri>' % self.download_uri(fn)
            msg += '<downloadFallbackUri>%s</downloadFallbackUri>' % self.download_fallback_uri(fn)
            logName = self.compressed_log_filename(fn)
            if os.path.exists(self.file_path(logName)):
                msg += self.xml_log(logName)
                includedLogs.append(logName)
            msg += '</package>'
            
            if distribVersion is None:
                distribVersion = self.version_from_filename(fn)

        # Any other logs we might want to include?
        for osName, osExt, osIdent in self.oses:
            for pkg in self.packages:
                logName = log_filename(pkg, osIdent)
                if os.path.exists(self.file_path(logName)) and logName not in includedLogs:
                    # Add an entry for this.
                    msg += '<package type="%s">' % self.package_type(logName)
                    msg += '<name>%s</name>' % self.packageName[pkg]
                    if self.version_from_filename(logName):
                        msg += '<version>%s</version>' % self.version_from_filename(logName)
                        if distribVersion is None:
                            distribVersion = self.version_from_filename(logName)
                    msg += '<platform>%s</platform>' % self.platId[osIdent]
                    msg += self.xml_log(logName)
                    msg += '</package>'
                            
        if distribVersion:
            msg += '<releaseNotes>%s</releaseNotes>' % self.release_notes_uri(distribVersion)
            msg += '<changeLog>%s</changeLog>' % self.changelog_uri(distribVersion)
        
        # Commits.
        chgFn = self.file_path('changes.xml')
        if os.path.exists(chgFn):
            msg += file(chgFn, 'rt').read()
                    
        return msg + '</build>'


def find_newest_event():
    newest = None
    for fn in os.listdir(config.EVENT_DIR):
        if fn[:5] != 'build': continue
        ev = Event(fn)
        bt = ev.timestamp()
        if newest is None or newest[0] < bt:
            newest = (bt, ev)

    if newest is None:
        return {'event':None, 'tag':None, 'time':time.time()}
    else:
        return {'event':newest[1], 'tag':newest[1].tag(), 'time':newest[0]}


def find_old_events(atLeastSecs):
    """Returns a list of Event instances."""
    result = []
    now = time.time()
    if not os.path.exists(config.EVENT_DIR): return result
    for fn in os.listdir(config.EVENT_DIR):
        if fn[:5] != 'build': continue
        ev = Event(fn)
        if now - ev.timestamp() >= atLeastSecs:
            result.append(ev)
    return result
    

def find_empty_events(baseDir=None):
    """Returns a list of build directory paths."""
    result = []
    if not baseDir: baseDir = config.EVENT_DIR
    print 'Finding empty subdirs in', baseDir
    for fn in os.listdir(baseDir):
        path = os.path.join(baseDir, fn)
        if os.path.isdir(path):
            # Is this an empty directory?
            empty = True
            for c in os.listdir(path):
                if c != '.' and c != '..':
                    empty = False
                    break
            if empty:
                result.append(path)
    return result


def events_by_time():
    builds = []
    for fn in os.listdir(config.EVENT_DIR):
        if fn[:5] == 'build':
            builds.append((Event(fn).timestamp(), Event(fn)))
    builds.sort()
    builds.reverse()
    return builds




# Copies all updated binaries to the distrib/products folder.
# Called from Visual Studio as a post-build step for all projects.

import sys, os, shutil

solDir = sys.argv[1]
buildType = sys.argv[2]


def copy_if_newer(src, dest):
    if os.path.exists(dest):
        srcTime = os.stat(src).st_mtime
        destTime = os.stat(dest).st_mtime
        if srcTime <= destTime: return    
    try:
        shutil.copyfile(src, dest)
        print 'DEPLOY :', os.path.basename(src), '->', os.path.abspath(dest)
    except IOError:
        pass


# Set up the destination.
dest = os.path.abspath(os.path.join(solDir, '..', 'distrib', 'products', 'bin'))
if not os.path.exists(dest): os.makedirs(dest)
if not os.path.exists(dest + '\\plugins'): os.makedirs(dest + '\\plugins')
if not os.path.exists(dest + '\\..\\data'): os.makedirs(dest + '\\..\\data')

# Update .pk3s.
copy_if_newer(os.path.join(solDir, 'doomsday.pk3'), os.path.join(dest, '..', 'data', 'doomsday.pk3'))
copy_if_newer(os.path.join(solDir, 'libdoom.pk3'), os.path.join(dest, '..', 'data', 'jdoom', 'libdoom.pk3'))
copy_if_newer(os.path.join(solDir, 'libheretic.pk3'), os.path.join(dest, '..', 'data', 'jheretic', 'libdoom.pk3'))
copy_if_newer(os.path.join(solDir, 'libhexen.pk3'), os.path.join(dest, '..', 'data', 'jhexen', 'libhexen.pk3'))
copy_if_newer(os.path.join(solDir, 'libdoom64.pk3'), os.path.join(dest, '..', 'data', 'jdoom64', 'libdoom64.pk3'))

targets = ['libcore', 'liblegacy', 'libshell', 'libgui', 'libappfw',
           'libdoomsday', 'client', 'server', 
           'plugins\*', 'tools\*', 'tools\shell\*', 'tests\*'] 

for target in targets:
    dirs = []
    if '*' in target:
        try:
            dirs += [target[:-1] + d + '\\' + buildType for d in os.listdir(os.path.join(solDir, target[:-1]))]
        except:
            pass
    else:
        dirs += [target + '\\' + buildType]
        
    for dir in dirs:
        d = os.path.join(solDir, dir)    
        if not os.path.exists(d): continue
               
        for fn in os.listdir(d):
            if fn[-4:] == '.exe':
                #print '- exe:', fn
                copy_if_newer(os.path.join(d, fn), os.path.join(dest, fn))
            elif fn[-4:] == '.dll':
                destPath = dest
                if 'plugins' in dir:
                    #print '- plugin dll:', fn
                    destPath += '\\plugins'
                #else:
                #    print '- dll:', fn
                copy_if_newer(os.path.join(d, fn), os.path.join(destPath, fn))



#!/usr/bin/env python
#-*- coding: UTF-8 -*-

""" This program loads a model with PyASSIMP, and display it.

It make a large use of shaders to illustrate a 'modern' OpenGL pipeline.

Based on:
 - pygame + mouselook code from http://3dengine.org/Spectator_%28PyOpenGL%29
 - http://www.lighthouse3d.com/tutorials
 - http://www.songho.ca/opengl/gl_transform.html
 - http://code.activestate.com/recipes/325391/
 - ASSIMP's C++ SimpleOpenGL viewer

Authors: Séverin Lemaignan, 2012-2013
"""
import sys

import logging
logger = logging.getLogger("pyassimp")
gllogger = logging.getLogger("OpenGL")
gllogger.setLevel(logging.WARNING)
logging.basicConfig(level=logging.INFO)

import OpenGL
OpenGL.ERROR_CHECKING=False
OpenGL.ERROR_LOGGING = False
#OpenGL.ERROR_ON_COPY = True
#OpenGL.FULL_LOGGING = True
from OpenGL.GL import *
from OpenGL.error import GLError
from OpenGL.GLU import *
from OpenGL.GLUT import *
from OpenGL.arrays import vbo
from OpenGL.GL import shaders

import pygame

import math, random
import numpy
from numpy import linalg

import pyassimp
from pyassimp.postprocess import *
from pyassimp.helper import *

class DefaultCamera:
    def __init__(self, w, h, fov):
        self.clipplanenear = 0.001
        self.clipplanefar = 100000.0
        self.aspect = w/h
        self.horizontalfov = fov * math.pi/180
        self.transformation = [[ 0.68, -0.32, 0.65, 7.48],
                               [ 0.73,  0.31, -0.61, -6.51],
                               [-0.01,  0.89,  0.44,  5.34],
                               [ 0.,    0.,    0.,    1.  ]]
        self.lookat = [0.0,0.0,-1.0]

    def __str__(self):
        return "Default camera"

class PyAssimp3DViewer:

    base_name = "PyASSIMP 3D viewer"

    def __init__(self, model, w=1024, h=768, fov=75):

        pygame.init()
        pygame.display.set_caption(self.base_name)
        pygame.display.set_mode((w,h), pygame.OPENGL | pygame.DOUBLEBUF)

        self.prepare_shaders()

        self.cameras = [DefaultCamera(w,h,fov)]
        self.current_cam_index = 0

        self.load_model(model)

        # for FPS computation
        self.frames = 0
        self.last_fps_time = glutGet(GLUT_ELAPSED_TIME)


        self.cycle_cameras()

    def prepare_shaders(self):

        phong_weightCalc = """
        float phong_weightCalc(
            in vec3 light_pos, // light position
            in vec3 frag_normal // geometry normal
        ) {
            // returns vec2( ambientMult, diffuseMult )
            float n_dot_pos = max( 0.0, dot(
                frag_normal, light_pos
            ));
            return n_dot_pos;
        }
        """

        vertex = shaders.compileShader( phong_weightCalc +
        """
        uniform vec4 Global_ambient;
        uniform vec4 Light_ambient;
        uniform vec4 Light_diffuse;
        uniform vec3 Light_location;
        uniform vec4 Material_ambient;
        uniform vec4 Material_diffuse;
        attribute vec3 Vertex_position;
        attribute vec3 Vertex_normal;
        varying vec4 baseColor;
        void main() {
            gl_Position = gl_ModelViewProjectionMatrix * vec4(
                Vertex_position, 1.0
            );
            vec3 EC_Light_location = gl_NormalMatrix * Light_location;
            float diffuse_weight = phong_weightCalc(
                normalize(EC_Light_location),
                normalize(gl_NormalMatrix * Vertex_normal)
            );
            baseColor = clamp(
            (
                // global component
                (Global_ambient * Material_ambient)
                // material's interaction with light's contribution
                // to the ambient lighting...
                + (Light_ambient * Material_ambient)
                // material's interaction with the direct light from
                // the light.
                + (Light_diffuse * Material_diffuse * diffuse_weight)
            ), 0.0, 1.0);
        }""", GL_VERTEX_SHADER)

        fragment = shaders.compileShader("""
        varying vec4 baseColor;
        void main() {
            gl_FragColor = baseColor;
        }
        """, GL_FRAGMENT_SHADER)

        self.shader = shaders.compileProgram(vertex,fragment)
        self.set_shader_accessors( (
            'Global_ambient',
            'Light_ambient','Light_diffuse','Light_location',
            'Material_ambient','Material_diffuse',
        ), (
            'Vertex_position','Vertex_normal',
        ), self.shader)

    def set_shader_accessors(self, uniforms, attributes, shader):
        # add accessors to the shaders uniforms and attributes
        for uniform in uniforms:
            location = glGetUniformLocation( shader,  uniform )
            if location in (None,-1):
                logger.warning('No uniform: %s'%( uniform ))
            setattr( shader, uniform, location )

        for attribute in attributes:
            location = glGetAttribLocation( shader, attribute )
            if location in (None,-1):
                logger.warning('No attribute: %s'%( attribute ))
            setattr( shader, attribute, location )


    def prepare_gl_buffers(self, mesh):

        mesh.gl = {}

        # Fill the buffer for vertex and normals positions
        v = numpy.array(mesh.vertices, 'f')
        n = numpy.array(mesh.normals, 'f')

        mesh.gl["vbo"] = vbo.VBO(numpy.hstack((v,n)))

        # Fill the buffer for vertex positions
        mesh.gl["faces"] = glGenBuffers(1)
        glBindBuffer(GL_ELEMENT_ARRAY_BUFFER, mesh.gl["faces"])
        glBufferData(GL_ELEMENT_ARRAY_BUFFER, 
                    mesh.faces,
                    GL_STATIC_DRAW)
        glBindBuffer(GL_ELEMENT_ARRAY_BUFFER,0)

    
    def load_model(self, path, postprocess = aiProcessPreset_TargetRealtime_MaxQuality):
        logger.info("Loading model:" + path + "...")

        if postprocess:
            self.scene = pyassimp.load(path, postprocess)
        else:
            self.scene = pyassimp.load(path)
        logger.info("Done.")

        scene = self.scene
        #log some statistics
        logger.info("  meshes: %d" % len(scene.meshes))
        logger.info("  total faces: %d" % sum([len(mesh.faces) for mesh in scene.meshes]))
        logger.info("  materials: %d" % len(scene.materials))
        self.bb_min, self.bb_max = get_bounding_box(self.scene)
        logger.info("  bounding box:" + str(self.bb_min) + " - " + str(self.bb_max))

        self.scene_center = [(a + b) / 2. for a, b in zip(self.bb_min, self.bb_max)]

        for index, mesh in enumerate(scene.meshes):
            self.prepare_gl_buffers(mesh)

        # Finally release the model
        pyassimp.release(scene)

        logger.info("Ready for 3D rendering!")

    def cycle_cameras(self):
        if not self.cameras:
            logger.info("No camera in the scene")
            return None
        self.current_cam_index = (self.current_cam_index + 1) % len(self.cameras)
        self.current_cam = self.cameras[self.current_cam_index]
        self.set_camera(self.current_cam)
        logger.info("Switched to camera <%s>" % self.current_cam)

    def set_camera_projection(self, camera = None):

        if not camera:
            camera = self.cameras[self.current_cam_index]

        znear = camera.clipplanenear
        zfar = camera.clipplanefar
        aspect = camera.aspect
        fov = camera.horizontalfov

        glMatrixMode(GL_PROJECTION)
        glLoadIdentity()

        # Compute gl frustrum
        tangent = math.tan(fov/2.)
        h = znear * tangent
        w = h * aspect

        # params: left, right, bottom, top, near, far
        glFrustum(-w, w, -h, h, znear, zfar)
        # equivalent to:
        #gluPerspective(fov * 180/math.pi, aspect, znear, zfar)
        glMatrixMode(GL_MODELVIEW)
        glLoadIdentity()


    def set_camera(self, camera):

        self.set_camera_projection(camera)

        glMatrixMode(GL_MODELVIEW)
        glLoadIdentity()

        cam = transform([0.0, 0.0, 0.0], camera.transformation)
        at = transform(camera.lookat, camera.transformation)
        gluLookAt(cam[0], cam[2], -cam[1],
                   at[0],  at[2],  -at[1],
                       0,      1,       0)

    def render(self, wireframe = False, twosided = False):

        glEnable(GL_DEPTH_TEST)
        glDepthFunc(GL_LEQUAL)


        glPolygonMode(GL_FRONT_AND_BACK, GL_LINE if wireframe else GL_FILL)
        glDisable(GL_CULL_FACE) if twosided else glEnable(GL_CULL_FACE)

        shader = self.shader

        glUseProgram(shader)
        glUniform4f( shader.Global_ambient, .4,.2,.2,.1 )
        glUniform4f( shader.Light_ambient, .4,.4,.4, 1.0 )
        glUniform4f( shader.Light_diffuse, 1,1,1,1 )
        glUniform3f( shader.Light_location, 2,2,10 )

        self.recursive_render(self.scene.rootnode, shader)


        glUseProgram( 0 )

    def recursive_render(self, node, shader):
        """ Main recursive rendering method.
        """

        # save model matrix and apply node transformation
        glPushMatrix()
        m = node.transformation.transpose() # OpenGL row major
        glMultMatrixf(m)

        for mesh in node.meshes:

            stride = 24 # 6 * 4 bytes

            diffuse = mesh.material.properties["diffuse"]
            if len(diffuse) == 3: diffuse.append(1.0)
            ambient = mesh.material.properties["ambient"]
            if len(ambient) == 3: ambient.append(1.0)

            glUniform4f( shader.Material_diffuse, *diffuse )
            glUniform4f( shader.Material_ambient, *ambient )

            vbo = mesh.gl["vbo"]
            vbo.bind()

            glEnableVertexAttribArray( shader.Vertex_position )
            glEnableVertexAttribArray( shader.Vertex_normal )

            glVertexAttribPointer(
                shader.Vertex_position,
                3, GL_FLOAT,False, stride, vbo
            )

            glVertexAttribPointer(
                shader.Vertex_normal,
                3, GL_FLOAT,False, stride, vbo+12
            )

            glBindBuffer(GL_ELEMENT_ARRAY_BUFFER, mesh.gl["faces"])
            glDrawElements(GL_TRIANGLES, len(mesh.faces) * 3, GL_UNSIGNED_INT, None)


            vbo.unbind()
            glDisableVertexAttribArray( shader.Vertex_position )

            glDisableVertexAttribArray( shader.Vertex_normal )


            glBindBuffer(GL_ELEMENT_ARRAY_BUFFER, 0)

        for child in node.children:
            self.recursive_render(child, shader)

        glPopMatrix()


    def loop(self):

        pygame.display.flip()
        pygame.event.pump()
        self.keys = [k for k, pressed in enumerate(pygame.key.get_pressed()) if pressed]

        glClear(GL_COLOR_BUFFER_BIT | GL_DEPTH_BUFFER_BIT)

        # Compute FPS
        gl_time = glutGet(GLUT_ELAPSED_TIME)
        self.frames += 1
        if gl_time - self.last_fps_time >= 1000:
            current_fps = self.frames * 1000 / (gl_time - self.last_fps_time)
            pygame.display.set_caption(self.base_name + " - %.0f fps" % current_fps)
            self.frames = 0
            self.last_fps_time = gl_time


        return True

    def controls_3d(self,
                    mouse_button=1, \
                    up_key=pygame.K_UP, \
                    down_key=pygame.K_DOWN, \
                    left_key=pygame.K_LEFT, \
                    right_key=pygame.K_RIGHT):
        """ The actual camera setting cycle """
        mouse_dx,mouse_dy = pygame.mouse.get_rel()
        if pygame.mouse.get_pressed()[mouse_button]:
            look_speed = .2
            buffer = glGetDoublev(GL_MODELVIEW_MATRIX)
            c = (-1 * numpy.mat(buffer[:3,:3]) * \
                numpy.mat(buffer[3,:3]).T).reshape(3,1)
            # c is camera center in absolute coordinates, 
            # we need to move it back to (0,0,0) 
            # before rotating the camera
            glTranslate(c[0],c[1],c[2])
            m = buffer.flatten()
            glRotate(mouse_dx * look_speed, m[1],m[5],m[9])
            glRotate(mouse_dy * look_speed, m[0],m[4],m[8])
            
            # compensate roll
            glRotated(-math.atan2(-m[4],m[5]) * \
                57.295779513082320876798154814105 ,m[2],m[6],m[10])
            glTranslate(-c[0],-c[1],-c[2])

        # move forward-back or right-left
        if up_key in self.keys:
            fwd = .1
        elif down_key in self.keys:
            fwd = -.1
        else:
            fwd = 0

        if left_key in self.keys:
            strafe = .1
        elif right_key in self.keys:
            strafe = -.1
        else:
            strafe = 0

        if abs(fwd) or abs(strafe):
            m = glGetDoublev(GL_MODELVIEW_MATRIX).flatten()
            glTranslate(fwd*m[2],fwd*m[6],fwd*m[10])
            glTranslate(strafe*m[0],strafe*m[4],strafe*m[8])

if __name__ == '__main__':
    if not len(sys.argv) > 1:
        print("Usage: " + __file__ + " <model>")
        sys.exit(2)

    app = PyAssimp3DViewer(model = sys.argv[1], w = 1024, h = 768, fov = 75)

    while app.loop():
        app.render()
        app.controls_3d(0)
        if pygame.K_f in app.keys: pygame.display.toggle_fullscreen()
        if pygame.K_TAB in app.keys: app.cycle_cameras()
        if pygame.K_ESCAPE in app.keys:
            break



#!/usr/bin/env python2.7
# This Python script will create a set of PK3 files that contain the files
# that Doomsday needs at runtime. The PK3 files are organized using the
# traditional data/ and defs/ structure.

import sys, os, os.path, zipfile

if len(sys.argv) < 2:
    print "Usage: %s pk3-target-dir" % sys.argv[0]
    print "(run in build/scripts/)"
    sys.exit(0)

# Check quiet flag.
quietMode = False
if '--quiet' in sys.argv:
    sys.argv.remove('--quiet')
    quietMode = True

deng_dir = os.path.join('..', '..')
target_dir = os.path.abspath(sys.argv[1])

class Pack:
    def __init__(self):
        self.files = [] # tuples

    def add_files(self, fileNamesArray):
        self.files += fileNamesArray

    def msg(self, text):
        if not quietMode: print text

    def create(self, name):
        full_name = os.path.join(target_dir, name)
        self.msg("Creating %s as %s..." % (os.path.normpath(name), os.path.normpath(full_name)))

        pk3 = zipfile.ZipFile(full_name, 'w', zipfile.ZIP_DEFLATED)

        for src, dest in self.files:
            full_src = os.path.join(deng_dir, src)
            # Is this a file or a folder?
            if os.path.isfile(full_src):
                # Write the file as is.
                self.msg("writing %s as %s" % (os.path.normpath(full_src), os.path.normpath(dest)))
                pk3.write(full_src, dest)
            elif os.path.isdir(full_src):
                # Write the contents of the folder recursively.
                def process_dir(path, dest_path):
                    self.msg("processing %s" % os.path.normpath(path))
                    for file in os.listdir(path):
                        real_file = os.path.join(path, file)
                        if file[0] == '.':
                            continue # Ignore these.
                        if os.path.isfile(real_file):
                            if not quietMode:
                                self.msg("writing %s as %s" % (os.path.normpath(real_file),
                                         os.path.normpath(os.path.join(dest_path, file))))
                            pk3.write(real_file, os.path.join(dest_path, file))
                        elif os.path.isdir(real_file):
                            process_dir(real_file,
                                        os.path.join(dest_path, file))
                process_dir(full_src, dest)

        # Write it out.
        print "Created %s (with %i files)." % (os.path.normpath(full_name), len(pk3.namelist()))
        pk3.close()

# First up, doomsday.pk3.
# Directory contents added recursively.
p = Pack()
p.add_files(
    [ ('client/data', 'data') ] )
p.create('doomsday.pk3')

# libdoom.pk3
p = Pack()
p.add_files(
    [ ('plugins/doom/defs', 'defs/jdoom'),
      ('plugins/doom/data/chex.mapinfo',           'data/jdoom/chex.mapinfo'),
      ('plugins/doom/data/doom1-share.mapinfo',    'data/jdoom/doom1-share.mapinfo'),
      ('plugins/doom/data/doom1-ultimate.mapinfo', 'data/jdoom/doom1-ultimate.mapinfo'),
      ('plugins/doom/data/doom1.mapinfo',          'data/jdoom/doom1.mapinfo'),
      ('plugins/doom/data/doom2-plut.mapinfo',     'data/jdoom/doom2-plut.mapinfo'),
      ('plugins/doom/data/doom2-tnt.mapinfo',      'data/jdoom/doom2-tnt.mapinfo'),
      ('plugins/doom/data/doom2-freedm.mapinfo',   'data/jdoom/doom2-freedm.mapinfo'),
      ('plugins/doom/data/doom2.mapinfo',          'data/jdoom/doom2.mapinfo'),
      ('plugins/doom/data/hacx.mapinfo',           'data/jdoom/hacx.mapinfo'),
      ('plugins/doom/data/conhelp.txt', 'data/jdoom/conhelp.txt'),
      ('plugins/doom/data/lumps', '#.basedata') ] )
p.create('libdoom.pk3')

# libheretic.pk3
p = Pack()
p.add_files(
    [ ('plugins/heretic/defs', 'defs/jheretic'),
      ('plugins/heretic/data/heretic-ext.mapinfo',   'data/jheretic/heretic-ext.mapinfo'),
      ('plugins/heretic/data/heretic-share.mapinfo', 'data/jheretic/heretic-share.mapinfo'),
      ('plugins/heretic/data/heretic.mapinfo',       'data/jheretic/heretic.mapinfo'),
      ('plugins/heretic/data/conhelp.txt', 'data/jheretic/conhelp.txt'),
      ('plugins/heretic/data/lumps', '#.basedata') ] )
p.create('libheretic.pk3')

# libhexen.pk3
p = Pack()
p.add_files(
    [ ('plugins/hexen/defs', 'defs/jhexen'),
      ('plugins/hexen/data/hexen-dk.mapinfo', 'data/jhexen/hexen-dk.mapinfo'),
      ('plugins/hexen/data/hexen.mapinfo',    'data/jhexen/hexen.mapinfo'),
      ('plugins/hexen/data/conhelp.txt',      'data/jhexen/conhelp.txt'),
      ('plugins/hexen/data/lumps', '#.basedata') ] )
p.create('libhexen.pk3')

# libdoom64.pk3
p = Pack()
p.add_files(
    [ ('plugins/doom64/defs', 'defs/jdoom64'),
      ('plugins/doom64/data/doom64.mapinfo', 'data/jdoom64/doom64.mapinfo'),
      ('plugins/doom64/data/conhelp.txt', 'data/jdoom64/conhelp.txt'),
      ('plugins/doom64/data/lumps', '#.basedata') ] )
p.create('libdoom64.pk3')



# -*- coding: iso-8859-1 -*-
# $Id$
# Snowberry: Extensible Launcher for the Doomsday Engine
#
# Copyright (C) 2004, 2005
#   Jaakko Keränen <jaakko.keranen@iki.fi>
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not: http://www.opensource.org/

## @file widgets.py User Interface Widgets
##
## This module contains a number of wrapper classes that wrap the real
## widget classes of wxWidgets.  This is done for purposes of
## convenience and event translation.  Events are converted into
## notifications that can be broadcasted using events.py.
##
## FormattedTabArea and TabArea could use a common base class.

import sys, os.path
import wx
import wx.html
import wx.lib.intctrl as intctrl
import wx.lib.scrolledpanel as scrolled
import wx.lib.fancytext as fancy
import host, events, ui, language, paths
import sb.profdb as pr
#import sb.confdb as st
import sb.widget.base


def uniConv(str):
    """Decode character encoding to Unicode.

    @param str  String in local encoding.
    
    @return Unicode string.
    """
    if host.isWindows():
        # Don't do Unicode on Windows.
        return str
    elif type(str) == unicode:
        return str
    else:
        return str.decode(host.getEncoding())


# The image list for all the icons used in the UI.
iconManager = None


def initIconManager():
    """Initialize a manager for all the icons used in the user
    interface.
    """
    wx.InitAllImageHandlers()

    global iconManager
    iconManager = IconManager(48, 48)


class IconManager:
    """The IconManager class provides a simpler identifier-based
    interface to wxWidgets image lists."""

    def __init__(self, width, height):
        """Construct a new icon manager.

        @param width Width of the icons.

        @param height Height of the icons.
        """
        self.imageList = wx.ImageList(width, height)

        # Identifiers of the icons in the list.
        self.bitmaps = []

    def getImageList(self):
        """Returns the wxImageList managed by this manager.

        @return A wxImageList object.
        """
        return self.imageList

    def get(self, identifier):
        """Finds the index number of the specified image.  If the
        image hasn't yet been loaded, it is loaded now.

        @param identifier Identifier of the image.
        """
        try:
            return self.bitmaps.index(identifier)

        except ValueError:
            # Load it now.  Images are affected by the localisation.
            if language.isDefined(identifier):
                imageName = language.translate(identifier)
            else:
                imageName = identifier
            fileName = paths.findBitmap(imageName)
            if len(fileName) == 0:
                # Fallback icon.
                fileName = paths.findBitmap('deng')
            bmp = wx.Bitmap(fileName)
            self.imageList.Add(bmp)
            self.bitmaps.append(identifier)
            return len(self.bitmaps) - 1
            
    def getBitmapPath(self, identifier, scaledSize=None):
        """Locates a bitmap and scales it, if necessary.
        
        @param identifier  Bitmap identifier.
        @param scaledSize  Tuple (width, height). If specified, determines 
                           the size of the image. If this does not match the
                           size of the existing image on disk, a new image 
                           file will be created.
        
        @return  Path of the image file.
        """
        if scaledSize is None:
            return paths.findBitmap(identifier)
        
        # A scaled size has been specified.
        # Let's generate the scaled name of the bitmap.
        ident = 'scaled%ix%i-' % scaledSize + identifier

        # Try to locate an existing copy of the image.
        imagePath = paths.findBitmap(ident)
        
        if imagePath != '':
            return imagePath

        # We have to generate it. First try loading the original image.
        originalPath = paths.findBitmap(identifier)
        if originalPath == '':
            # No such image.
            return ''

        # Scale and save. PNG should work well with all source images.
        image = wx.Image(originalPath)
        originalSize = (image.GetWidth(), image.GetHeight())
        if originalSize[0] == scaledSize[0] and originalSize[1] == scaledSize[1]:
            # No need to scale.
            return originalPath
        if scaledSize[0] < originalSize[0] or scaledSize[1] < originalSize[1]:
            # Upscale first to cause some extra blurring.
            image.Rescale(originalSize[0]*2, originalSize[1]*2, wx.IMAGE_QUALITY_HIGH)
        image.Rescale(scaledSize[0], scaledSize[1], wx.IMAGE_QUALITY_HIGH)
        imagePath = os.path.join(paths.getUserPath(paths.GRAPHICS), ident + '.png')
        image.SaveFile(imagePath, wx.BITMAP_TYPE_PNG)
                              
        return imagePath


class Line (sb.widget.base.Widget):
    """A static horizontal divider line."""

    def __init__(self, parent):
        """Construct a new divider line.

        @param parent The parent panel.
        """
        sb.widget.base.Widget.__init__(self, 
            wx.StaticLine(parent, -1, style=wx.LI_HORIZONTAL))


class Image (sb.widget.base.Widget):
    """A static image widget.  The user will not be able to interact
    with instances of the Image widget."""

    def __init__(self, parent, imageName):
        """Construct a new Image widget.

        @param imageName Base name of the image file.  It must be
        located in one of the graphics directories.
        """
        # Create a new panel so we can control the background color.
        self.panel = wx.Panel(parent, -1)
        self.panel.SetBackgroundColour(wx.WHITE)
        sb.widget.base.Widget.__init__(self, self.panel)

        # See if we can find the right image.
        bmp = wx.Image(paths.findBitmap(imageName)).ConvertToBitmap()

        self.bitmap = wx.StaticBitmap(self.panel, -1, bmp)
        self.panel.SetMinSize(bmp.GetSize())

    def setImage(self, imageName):
        """Change the image displayed in the widget.

        @param imageName Base name of the image file.  It must be
        located in one of the graphics directories.
        """
        bmp = wx.Image(paths.findBitmap(imageName)).ConvertToBitmap()
        self.bitmap.Freeze()
        self.bitmap.SetBitmap(bmp)
        self.panel.SetMinSize(bmp.GetSize())
        self.bitmap.Thaw()
        self.panel.Refresh()



#!/usr/bin/env python2.7

import os, sys
import conhelp

dengRoot = os.path.abspath(os.path.join(
    os.path.dirname(os.path.abspath(sys.argv[0])), '..', '..'))
    
conHelp = os.path.join(dengRoot, 'build', 'scripts', 'conhelp.py')

os.chdir(os.path.join(dengRoot, 'doc'))

print "Engine help"
cpHelpTxt = os.path.join(dengRoot, 'net.dengine.base.pack', 'helpstrings.txt')
conhelp.makeHelp(cpHelpTxt, ['engine'])

panel = conhelp.amethyst(file(os.path.join('engine', 'controlpanel.ame')).read())

file(cpHelpTxt, 'a').write(panel)

print "libdoom help"
conhelp.makeHelp(
    os.path.join(dengRoot, 'plugins', 'doom', 'data', 'conhelp.txt'),
    ['libcommon', 'libdoom'])
    
print "libheretic help"
conhelp.makeHelp(
    os.path.join(dengRoot, 'plugins', 'heretic', 'data', 'conhelp.txt'),
    ['libcommon', 'libheretic'])
    
print "libhexen help"
conhelp.makeHelp(
    os.path.join(dengRoot, 'plugins', 'hexen', 'data', 'conhelp.txt'),
    ['libcommon', 'libhexen'])



# -*- coding: iso-8859-1 -*-
# $Id$
# Snowberry: Extensible Launcher for the Doomsday Engine
#
# Copyright (C) 2004, 2006
#   Jaakko Keränen <jaakko.keranen@iki.fi>
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not: http://www.opensource.org/

## @file confdb.py Configuration Database
##
## When the module is initialized, the configuration files are read
## from the configuration directories.  We'll expect to find both
## components and settings.

# Import modules required for initialization.
import os, re, string
import paths
import events    
import logger
import host
import language
import cfparser
import conf


# Dictionary for system configuration (fonts, buttons sizes).
sysConfig = {}

# Dictionary for all known components.
allComponents = {}

# Dictionary for all normal settings (no system settings).
allSettings = {}

# Array of setting identifiers used to select components.
# [number of settings, number of components, array]
# Recalculated when the first two change.
compSettings = [0, 0, []]

# Dictionary for system settings.
systemSettings = {}


def _newComponent(component):
    """Add a new Component to the allComponents dictionary."""
    # Dictionaries are good because they don't allow duplicate IDs.
    allComponents[component.getId()] = component
    

def getComponent(id):
    """Returns the component with the specified identifier."""
    return allComponents[id]


def getComponents():
    """Returns an array of all the defined components."""
    return allComponents.values()


def getGameComponents():
    """Returns an array of all the game components."""
    games = []
    for c in getComponents():
        if c.getId()[:5] == 'game-':
            games.append(c)
    return games


def _newSetting(setting):
    """Add a new Setting to the allSettings dictionary.

    @param setting A Setting object.

    @return The same Setting object that was given as a parameter.
    """
    # Dictionaries are good because they don't allow duplicate IDs.
    allSettings[setting.getId()] = setting
    return setting


def _newSystemSetting(systemSetting):
    """Add a new Setting to the systemSettings dictionary.

    @param systemSetting A Setting object.

    @return The same Setting object that was given as a parameter.
    """
    # Dictionaries are good because they don't allow duplicate IDs.
    systemSettings[systemSetting.getId()] = systemSetting
    sysConfig[systemSetting.getId()] = systemSetting.getDefault()
    return systemSetting


def isSettingDefined(id):
    """Determines whether a setting with a specific identifier has been 
    defined."""
    return allSettings.has_key(id)


def getSetting(id):
    """Returns the setting with the specified identifier."""
    return allSettings[id]


def getSystemSetting(id):
    """Returns the system setting with the specified identifier."""
    return systemSettings[id]


def getSettings(func=None):
    """Returns a filtered list of settings.

    @param func If this function returns True, the setting will be
    included in the returned array.
    """
    filtered = []
    for s in allSettings.values():
        if func == None or func(s):
            filtered.append(s)
    return filtered
    
    
def getSystemSettings(func=None):
    """Returns a filtered list of system settings.

    @param func If this function returns True, the system setting will be
    included in the returned array.
    """
    return [s for s in systemSettings.values() if func is None or func(s)]   


def readConfigFile(fileName):
    """Reads a configuration file from the specified path.  The file
    must be a .conf file that contains either settings or components,
    or both."""
    p = cfparser.FileParser(fileName)

    # We'll collect all the elements into this array.
    elements = []
    
    try:
        # Loop through all the elements in the config file.
        while True:
            e = p.get()

            if e.isKey() and e.getName() == 'required-platform':
                # This configuration file is specific to a particular
                # platform.
                if (e.getValue() == 'windows' and not host.isWindows()) or \
                   (e.getValue() == 'mac' and not host.isMac()) or \
                   (e.getValue() == 'unix' and not host.isUnix()):
                    # Look, we ran out early.
                    raise cfparser.OutOfElements

            # Config files may contain blocks that group similar
            # settings together.
            if e.isBlock() and e.getType() == 'group':
                # Add everything inside the group into the elements.
                # A group can only contain settings.
                for sub in e.getContents():
                    # There can only be blocks in a group.
                    if not sub.isBlock():
                        continue
                    # Add the 'group' key element to the setting.
                    sub.add(cfparser.KeyElement('group', e.getName()))
                    elements.append(sub)
            else:
                elements.append(e)

    except cfparser.OutOfElements:
        # Parsing was completed.
        pass

    except Exception, exception:
        # Log an error message.
        import traceback
        traceback.print_exc()
        logger.add(logger.HIGH, 'error-read-config-file', fileName,
                  str(exception))

    # Process all the elements we got from the configuration files.
    for e in elements:

        # There should only be blocks in the array.
        if not e.isBlock():
            continue

        processSettingBlock(e)


def processSettingBlock(e):
    """Process a Block that contains the definition of a setting.

    @param e A cfparser.BlockElement object.
    """
    if not e.isBlock():
        return

    # First check for generic configuration options.  These can't be
    # changed in the user interface.
    if e.getType() == 'appearance' or e.getType() == 'configure':
        # Insert the keys into the system configuration dictionary.
        for key in e.getContents():
            if e.getName() == 'addon-path':
                # A custom addon path.
                paths.addAddonPath(key.getValue())
            else:
                # Regular sysConfig string.
                fullName = e.getName() + '-' + key.getName()
                sysConfig[fullName] = key.getValue()
        return
        
    # Check for languages.
    if e.getType() == 'language':
        language.processLanguageBlock(e)
        return
    
    setting = None

    # Each block represents either a component or a setting.
    if e.getType() == 'component':
        # Create a new component.
        _newComponent(conf.Component(e.getName(),
                                     e.findValue('setting'),
                                     e.findValue('option')))

    elif e.getType() == 'toggle':
        setting = conf.ToggleSetting(e.getName(),
                                     e.findValue('option'),
                                     e.findValue('default'),
                                     e.findValue('option-inactive'))

    elif e.getType() == 'implicit':
        setting = conf.Setting(e.getType(), e.getName(), e.findValue('option'))

    elif e.getType() == 'range':
        setting = conf.RangeSetting(e.getName(),
                                    e.findValue('option'),
                                    e.findValue('min'),
                                    e.findValue('max'),
                                    e.findValue('default'),
                                    e.findValue('suffix'))

    elif e.getType() == 'slider':
        setting = conf.SliderSetting(e.getName(),
                                     e.findValue('option'),
                                     e.findValue('min'),
                                     e.findValue('max'),
                                     e.findValue('step'),
                                     e.findValue('default'),
                                     e.findValue('suffix'))

    elif e.getType() == 'text':
        setting = conf.TextSetting(e.getName(),
                                   e.findValue('option'),
                                   e.findValue('default'))

    elif e.getType() == 'file':
        # Is there a list of allowed file types?
        allowedTypes = []
        typeBlock = e.find('types')
        if typeBlock and typeBlock.isBlock() and \
           typeBlock.getType() == 'allowed':
            for typeKey in typeBlock.getContents():
                allowedTypes.append((typeKey.getName(), typeKey.getValue()))
        
        setting = conf.FileSetting(e.getName(),
                                   e.findValue('option'),
                                   e.findValue('must-exist'),
                                   allowedTypes,
                                   e.findValue('default'))

    elif e.getType() == 'folder':
        setting = conf.FolderSetting(e.getName(),
                                     e.findValue('option'),
                                     e.findValue('must-exist'),
                                     e.findValue('default'))
                                      
    elif e.getType() == 'choice':
        # Identifiers of the choices.
        altsElement = e.find('alts')
        if altsElement:
            alts = altsElement.getContents()
        else:
            alts = []
            
        # Command line parameters of each choice.
        optsElement = e.find('opts')
        if optsElement:
            opts = optsElement.getContents()
        else:
            opts = []
                
        # Check what to do if existing choices are already defined.
        existing = e.findValue('existing')
        if existing == 'merge' and isSettingDefined(e.getName()):
            oldSetting = getSetting(e.getName())
            oldSetting.merge(alts, opts)
        else:
            # A totally new setting.
            setting = conf.ChoiceSetting(e.getName(),
                                         e.findValue('option'),
                                         alts, opts,
                                         e.findValue('default'))
                                     
    if setting:    
        # Any required values?
        req = e.find('equals')
        if req and req.getType() == 'require':
            for r in req.getContents():
                setting.addRequiredValue(r.getName(), r.getValue())
            
        # Grouping information?
        if e.find('group'):
            setting.setGroup(e.findValue('group'))
        if e.find('subgroup'):
            setting.setSubGroup(e.findValue('subgroup'))

        # Automatically append value after option on the command line?
        if e.find('value-after-option'):
            setting.setValueAfterOption(e.findValue('value-after-option'))
            
        # Any required components or addons?
        setting.addRequiredComponent(e.findValue('component'))
        setting.addRequiredAddon(e.findValue('addon'))
            
        # Add the setting to the dictionary of all settings.
        _newSetting(setting)
        

def getCompatibleSettings(profile):
    """Compose a list of all the settings that are compatible with the
    given profile.  Compatibility is determined based on the
    components and addons required by the settings.

    @param profile A profiles.Profile object.

    @return An array of Setting objects.
    """
    compatible = []

    # The profile's components and addons that have been selected for
    # loading.
    profileComponents = profile.getComponents()
    profileAddons = profile.getAddons()

    import sb.profdb

    # Iterate through all the available settings.
    for setting in allSettings.values():

        # Check the required components.  All settings are compatible
        # with the Defaults profile.
        if profile is not sb.profdb.getDefaults():

            for requirement in setting.getRequiredComponents():
                if requirement not in profileComponents:
                    # This setting is incompatible.
                    continue

            for requirement in setting.getRequiredAddons():
                if requirement not in profileAddons:
                    # This setting is incompatible.
                    continue

        # This appears to be a compatible setting.
        compatible.append(setting)

    return compatible


def getComponentSettings():
    """Compile an array of settings that are used to select components.

    @return An array of setting identifiers.
    """
    global compSettings

    # Do we need to look again?
    if len(allSettings) != compSettings[0] or \
       len(allComponents) != compSettings[1]:
        # Look again.
        compSettings[0] = len(allSettings)
        compSettings[1] = len(allComponents)
        compSettings[2] = []

        for s in allSettings.values():
            for c in allComponents.values():
                if c.getSetting() == s.getId():
                    compSettings[2].append(s.getId())
                    break

    return compSettings[2]


def haveSettingsForAddon(addon):
    """Checks if there are any settings for the specified addon. This
    only affects non-implicit settings, i.e. settings that the user
    can manipulate.

    @param addon Addon identifier.
    """
    for s in allSettings.values():
        if s.getType() == 'implicit':
            continue
        if addon in s.getRequiredAddons():
            return True

    return False


def _earliestInGroup(groupName):
    """Returns the earliest setting order number in the group."""
    earliest = conf.creationOrder

    for s in allSettings.values():
        if s.getGroup() == groupName:
            if earliest == None or s.getOrder() < earliest:
               earliest = min(s.getOrder(), earliest)

    return earliest


def getSettingGroups(addonIdentifier=None):
    """Compose a list of all groups that appear in the defined
    settings.  The list won't include groups that are used by settings
    that are associated with addons.

    @param addonIdentifier If not None, only finds the groups that are
    associated with the given addon.

    @return An array of identifiers.
    """
    groups = []

    for s in allSettings.values():
        if (addonIdentifier == None and len(s.getRequiredAddons()) or
            addonIdentifier != None and
            addonIdentifier not in s.getRequiredAddons()):
            # Settings associated with addons are ignored here.
            continue
        if s.getGroup() and s.getGroup() not in groups:
            groups.append(s.getGroup())

    groups.sort(lambda a, b: cmp(_earliestInGroup(a),
                                 _earliestInGroup(b)))

    return groups


def isDefined(id):
    """Checks if a certain system configuration setting has been defined.

    @param id Identifier of the system configuration setting.

    @return True, if the setting has been defined.
    """
    return sysConfig.has_key(id)
    
    
def undefine(id):
    """Undefines a certain system configuration setting.
    
    @param id Identifier of the system configuration setting."""
    if isDefined(id):
        del sysConfig[id]


def getSystemString(id):
    """Return the value of a system configuration setting.  These
    include things such as fonts and button sizes."""
    return sysConfig[id]


def getSystemInteger(id, defaultValue=None):
    """Return the value of a system configuration setting.  These
    include things such as fonts and button sizes."""
    try:
        return int(getSystemString(id))
    except:
        if defaultValue != None:
            return defaultValue
        return 0
    

def getSystemBoolean(id):
    """Return the boolean value of a system configuration setting."""
    return isDefined(id) and getSystemString(id) == 'yes'        


def readConfigPath(path):
    """Load all the config files in the specified directory.

    @param path Directory path.
    """
    try:
        for name in os.listdir(path):
            # We'll load all the files with the .conf extension.
            if paths.hasExtension('conf', name):
                # Load this configuration file.
                readConfigFile(os.path.join(path, name))
    except OSError:
        # Path does not exist.
        pass


def handleNotify(event):
    """Handle notifications.

    @param event An events.Notify event.
    """
    if event.hasId('quit'):
        # The system settings aren't part of any profile, so they need
        # to be saved separately.
        saveSystemConfigSettings()

    elif event.hasId('widget-edited'):
        # When system settings are modified, we immediately update the
        # sysConfig dictionary here.
        if isDefined(event.getWidget()):
            sysConfig[event.getWidget()] = event.getValue()


def saveSystemConfigSettings():
    """Write the values of the system config settings into a
    configuration file called <tt>system.conf</tt>."""

    fileName = os.path.join(paths.getUserPath(paths.CONF), 'system.conf')
    try:
        f = file(fileName, 'w')
        f.write('# This file is generated automatically.\n')
        f.write('configure snowberry ( previous-version = %s )\n' %
                getSystemString('snowberry-version'))

        for s in systemSettings:
            parts = s.split('-')
            f.write('configure %s ( ' % parts[0])
            f.write('%s = ' % string.join(parts[1:], '-'))
            f.write('%s )\n' % getSystemString(s))

    except: 
        # Paths not saved.
        import traceback
        traceback.print_exc()
        
        
def checkUpgrade():
    """Check the current version against the previous version."""
    
    if not isDefined('snowberry-previous-version'):
        previousVersion = '1.1 or earlier'
    else:
        previousVersion = getSystemString('snowberry-previous-version')

    currentVersion = getSystemString('snowberry-version')

    if previousVersion != currentVersion:
        if previousVersion == '1.1 or earlier':
            # Reset some settings to their default values.
            sysConfig['main-hide-title'] = 'yes'
            sysConfig['main-hide-help'] = 'yes'
            sysConfig['profile-large-icons'] = 'no'
            undefine('main-split-position')
            undefine('main-profile-split-position')
            undefine('main-width')
            undefine('main-height')
                        

def init():
    """Initialize the module."""
    
    # Built-in settings.
    allLangs = language.getLanguages()
    lang = conf.ChoiceSetting('language', '', allLangs, [], 'english')
    lang.setGroup('general-options')
    lang.setSorted(True)
    _newSetting(lang)

    quitLaunch = conf.ToggleSetting('quit-on-launch', '', 'yes', '')
    quitLaunch.setGroup('general-options')
    _newSetting(quitLaunch)

    # System settings.
    tog = conf.ToggleSetting('main-hide-title', '', 'yes', '')
    _newSystemSetting(tog)
    tog = conf.ToggleSetting('main-hide-help', '', 'yes', '')
    _newSystemSetting(tog)
    tog = conf.ToggleSetting('summary-profile-change-autoselect', '', 'no', '')
    _newSystemSetting(tog)
    tog = conf.ToggleSetting('profile-large-icons', '', 'no', '')
    _newSystemSetting(tog)
    tog = conf.ToggleSetting('profile-hide-buttons', '', 'no', '')
    _newSystemSetting(tog)
    tog = conf.ToggleSetting('profile-minimal-mode', '', 'no', '')
    _newSystemSetting(tog)

    # Load all .conf files.
    for path in paths.listPaths(paths.CONF, False):
        readConfigPath(path)

    # Any custom paths?
    if isDefined('doomsday-runtime'):
        paths.setCustomPath(paths.RUNTIME, getSystemString('doomsday-runtime'))

    # Check against previous version to detect upgrades.
    checkUpgrade()

    # Listen for the quit notification so we can save some settings.
    events.addNotifyListener(handleNotify, ['quit', 'widget-edited'])


# Initialize module.    
init()    



# -*- coding: iso-8859-1 -*-
# $Id$
# Snowberry: Extensible Launcher for the Doomsday Engine
#
# Copyright (C) 2004, 2006
#   Jaakko Keränen <jaakko.keranen@iki.fi>
#   Veikko Eeva
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not: http://www.opensource.org/

## @file aodb.py Addon Database
##
## This module handles the functionality related to addons including,
## for instance, database administration, installing and un-installing
## addons.

import os, re, string, zipfile, shutil, struct
import logger, events, paths, language, cfparser
import addon as ao
import sb.profdb

# This is a list of all the categories that include exclusions regarding
# the other categories. That is, some addons cannot be included in Doomsday
# Engine launch parameters at the same time.
categories = []
rootCategory = None

# The addon dictionary and the metadata cache.  Maps addon identifiers
# to Addon objects.
addons = {}


class Category:
    """The category tree is built automatically based on the loaded
    addons."""

    def __init__(self, id, parent=None):
        self.id = id
        self.parent = parent

    def getId(self):
        return self.id

    def getLongId(self):
        """The long identifier includes the identifiers of the parent
        categories."""
        
        longId = self.id
        p = self.parent
        while p:
            longId = p.id + '-' + longId
            p = p.parent
        return 'category' + longId

    def getParent(self):
        """Returns the parent category.

        @param A Category object.
        """
        return self.parent

    def getName(self):
        """Returns the user-visible name of the category.

        @param A string.  The translated name of the category.
        """
        return language.translate(self.getLongId())

    def isAncestorOf(self, subCategory):
        """Checks if this Category is an ancestor of another category.
        Also true if the categories are the same object.

        @param subCategory The potential subcategory.

        @return True, if subCategory is a descendant of this category.
        """
        i = subCategory
        while i:
            if i is self:
                return True
            i = i.getParent()
        return False

    def getPath(self):
        """Returns the category in the path notation: aaa/bbb/ccc.

        @return A string.
        """
        path = ''
        iter = self
        while iter.getParent():
            path = '/' + iter.getId() + path
            iter = iter.getParent()
        return path


def _getLatestModTime(startPath):
    """Descends recursively into the subdirectories in startPath to
    find out which file has the latest modification time.

    @return The latest modification time.
    """
    latest = 0

    for f in os.listdir(startPath):
        path = os.path.join(startPath, f)
        if os.path.isdir(path):
            # Check which is the latest file in this subdirectory.
            latest = max(latest, _getLatestModTime(path))
        else:
            latest = max(latest, os.path.getmtime(path))

    return latest


def _newCategory(identifier, parent):
    """Append a new Category object into the array of categories.
    This is only done if the specified category does not already
    exist.  This shouldn't be called directly.  Call _getCategory()
    instead.

    @param category A Category object.

    @return The real Category object that must be used.  This may be
    different than the provided Category object, since the category
    may already exist.
    """
    for c in categories:
        if c.getId() == identifier and c.getParent() is parent:
            # This already exists.
            return c

    # Construct a new category object.
    c = Category(identifier, parent)
    categories.append(c)
    return c


def _getCategory(name, prefix=None):
    """Finds the Category object that corresponds the specified
    category path.  If the object doesn't exist yet, it will be
    created.  The categories form a tree hierarchy.

    @param The category path.  Separate identifiers with slashes.
    """
    trimmed = name.lower()

    # Remove any spaces and a trailing/initial slash.
    while trimmed.startswith('/'):
        trimmed = trimmed[1:]
    while trimmed.endswith('/'):
        trimmed = trimmed[:-1]

    # Get rid of all duplicate slashes.
    while True:
        cleaned = trimmed.replace('//', '/')
        if cleaned == trimmed:
            break
        trimmed = cleaned

    if len(trimmed) == 0:
        return rootCategory

    # Get the category identifiers in the path, stripped of
    # whitespace.
    idents = map(lambda a: a.strip(), trimmed.split('/'))

    cat = rootCategory

    # Build the category objects in the path, if they don't yet exist.
    for i in idents:
        # Create the new categories and make sure it's in the category
        # tree.
        newCat = _newCategory(i, cat)
        cat = newCat

    return cat


def getAddonExtensions():
    """Returns a list of all possible file name extensions for addons."""
    return ao.EXTENSIONS


def getCategory(longId):
    """Finds a Category by its long identifier.
    
    @param longId  Long category identifier.
    
    @return Category, or None.
    """
    for c in categories:
        if c.getLongId() == longId:
            return c
    return None


def getRootCategory():
    """Returns the root category.

    @return A Category object."""
    return rootCategory


def getSubCategories(parent):
    """Return the sub-categories of a parent category.

    @param parent A Category object.

    @return An array of Category objects.
    """
    return [c for c in categories if c.getParent() is parent]


def getAvailableAddons(profile):
    """Get all addons compatible with the profile.

    @return An array of Addon objects.
    """
    import sb.profdb
    return [a for a in addons.values()
            if profile is sb.profdb.getDefaults() or a.isCompatibleWith(profile)]


def getAddonCount():
    """Returns the number of addons in the addon database."""
    return len(addons)


def getAddons(func=None):
    """Return the addons in a given category.

    @param func If this function returns True, the addon will be
    returned.

    @return An array of Addon objects.
    """
    return [a for a in addons.values() if func == None or func(a)]


def sortIdentifiersByName(addonIdentifiers):
    """Sort a list of addon identifiers based on their visible names
    in the current language."""
    addonIdentifiers.sort(lambda a, b: cmp(language.translate(a),
                                           language.translate(b)))


def exists(addonId):
    return addons.has_key(addonId)


def get(addonId):
    """Returns the specified addon.

    @param addonId A string identifier.

    @return An Addon object.
    """
    return addons[addonId]


def install(sourceName):
    """Installs the specified addon into the user's addon database.

    @param fileName Full path of the addon to install.

    @throw AddonFormatError If the addon is not one of the recognized
    formats, this exception is raised.

    @return Identifier of the installed addon.

    @throws Exception The installation failed.
    """
    isManifest = paths.hasExtension('manifest', sourceName)
    
    if isManifest:
        # Manifests are copied to the manifest directory.
        destPath = paths.getUserPath(paths.MANIFESTS)
    else:
        destPath = paths.getUserPath(paths.ADDONS)

    destName = os.path.join(destPath,
                            os.path.basename(sourceName))
    
    try:
        # Directories must be copied as a tree.
        if os.path.isdir(sourceName):
            shutil.copytree(sourceName, destName)
        else:
            shutil.copy2(sourceName, destName)

        # Load the new addon.
        if not isManifest:
            identifier = load(destName)
        else:
            identifier = loadManifest(destName)

        # Now that the addon has been installed, save the updated cache.
        writeMetaCache()

        # Send a notification of the new addon.
        notify = events.Notify('addon-installed')
        notify.addon = identifier
        events.send(notify)

    except Exception, x:
        # Unsuccessful.
        raise x

    return identifier


def isUninstallable(identifier):
    """Determines whether the addon is uninstallable.
    @param identifier  Addon identifier.
    @return Bool."""
    return exists(identifier) and get(identifier).getBox() == None
    

def uninstall(identifier, doNotify=False):
    """Uninstall an addon.  Uninstalling an addon causes a full reload
    of the entire addon database.
    
    @param identifier  Identifier of the addon to uninstall.
    @param doNotify  True, if a notification should be sent.
    """
    addon = get(identifier)
    path = addon.getSource()
    destPath = os.path.join(paths.getUserPath(paths.UNINSTALLED),
                            os.path.basename(path))

    if os.path.exists(destPath):
        # Simply remove any existing addons with the same name.
        # TODO: Is this wise? Rename old.
        if os.path.isdir(destPath):
            shutil.rmtree(destPath)
        else:
            os.remove(destPath)

    shutil.move(path, destPath)

    # Is there a manifest to uninstall?
    manifest = addon.getId() + '.manifest'
    manifestFile = os.path.join(paths.getUserPath(paths.MANIFESTS), manifest)
    if os.path.exists(manifestFile):
        # Move it to the uninstalled folder.
        destPath = os.path.join(paths.getUserPath(paths.UNINSTALLED),
                                manifest)
        if os.path.exists(destPath):
            os.remove(destPath)
        shutil.move(manifestFile, destPath)

    # Mark as uninstalled.
    addon.uninstall()

    # Detach from all profiles.
        
    for p in sb.profdb.getProfiles():
        p.dontUseAddon(identifier)

    # Send a notification.
    if doNotify:
        notify = events.Notify('addon-uninstalled')
        notify.addon = identifier
        events.send(notify)   


def formIdentifier(fileName):
    """Forms the identifier of the addon based on the addon's file name.

    @param fileName  File name of the addon.

    @return  Tuple: (identifier, effective extension)
    """
    baseName = os.path.basename(fileName)
    found = re.search("(?i)(.*)\.(" + \
                      string.join(ao.EXTENSIONS, '|') + ")$", baseName)
    if not found:
        return ('', '')
    
    identifier = found.group(1).lower()
    extension = found.group(2).lower()
    if extension == 'addon':
        extension = 'bundle'

    # The identifier also includes the type of the addon.
    identifier += "-" + extension
    
    # Replace spaces with dashes.
    identifier = identifier.replace(' ', '-')
    
    # Replace reserved characters with underscores.
    identifier = identifier.replace('(', '_')
    identifier = identifier.replace(')', '_')
    identifier = identifier.replace('>', '_')
    identifier = identifier.replace('<', '_')
    identifier = identifier.replace(',', '_')

    return (identifier, extension)


class LoadError (Exception):
    """Exception that is raised when the loading of an addon fails."""
    def __init__(self, msg=''):
        Exception.__init__(self, msg)


def load(fileName, containingBox=None):
    """Loads the specific addon.  If its metadata hasn't yet been
    cached, it will be retrieved and stored into the data cache.  All
    addons in the system and user databases are loaded on program
    startup.

    @param fileName Full path of the addon to load.

    @param containingBox Optionally the BoxAddon that contains the
    loaded addon.

    @return The identifier of the loaded addon.
    """
    # Form the identifier of the addon.
    identifier, extension = formIdentifier(fileName)

    addon = None
    
    # Construct a new Addon object based on the file name extension.
    if extension == 'box':
        addon = ao.BoxAddon(identifier, fileName)
    elif extension == 'bundle':
        addon = ao.BundleAddon(identifier, fileName)
    elif extension == 'pk3' or extension == 'zip':
        addon = ao.PK3Addon(identifier, fileName)
    elif extension == 'wad':
        addon = ao.WADAddon(identifier, fileName)
    elif extension == 'ded':
        addon = ao.DEDAddon(identifier, fileName)
    elif extension == 'deh':
        addon = ao.DehackedAddon(identifier, fileName)
    elif extension == 'lmp':
        addon = ao.LumpAddon(identifier, fileName)
    else:
        raise LoadError('unknown format')

    # Set the container.
    if containingBox:
        addon.setBox(containingBox)

    # Check against the current data in the cache.  Is it recent
    # enough?
    lastModified = addon.getLastModified()

    if (not addons.has_key(identifier) or
        addons[identifier].getLastModified() != lastModified):

        # The metadata needs to be updated.
        addon.readMetaData()

        # This Addon object contains the latest metadata.
        addons[identifier] = addon

    # If the loaded addon is a box, also load the contents.
    if extension == 'box':
        addon.loadContents()

    # TODO: Define the strings of this addon in the language.

    return identifier


def loadManifest(fileName):
    """Manifests contain metadata for other addons.  A manifest may
    augment or replace the metadata of an existing addon, or define an
    entirely new addon.

    @param fileName  Path of the manifest file.

    @return  Identifier of the addon the manifest is associated with.
    """
    identifier = paths.getBase(fileName)

    if exists(identifier):
        a = get(identifier)
    else:
        # Create a new addon.
        a = ao.Addon(identifier, fileName)
        addons[identifier] = a

    # The manifest contains metadata configuration.
    try:
        a.parseConfiguration(file(fileName).read())
        
    except Exception, x:
        logger.add(logger.HIGH, 'error-read-manifest', fileName, str(x))

    return identifier


def readMetaCache():
    """The metadata cache contains the metadata extracted from all the
    installed addons.  The cache is stored in the user's addon
    database directory.  The entries whose source file is missing will
    be ignored.
    """


def writeMetaCache():
    """Write the metadata cache to disk.
    """


def findConflicts(roster, profile):
    """Find conflicts in the given list of addon identifiers.  This
    function only finds any conflicts that exist, and then returns a
    tuple describing what must be done to resolve the conflict.  This
    is called repeatedly by the launcher plugin until no more
    conflicts are found.

    @param roster An array of Addon objects.  The addons with the
    highest priority are in the end of the list (first in the load
    order; overrides the addons loaded earlier).

    @param profile The profile that is used with the addons.  The
    values of settings are taken for here.

    @return An empty list if no conflicts were found.  Otherwise a list
    of conflict reports (as tuples).
    """
    rosterSize = len(roster)

    result = []

    # Get the keywords defined by the profile's values.
    settingKeywords = profile.getKeywords()
    
    # Collect a list of all the keywords that are provided and
    # offered.
    allKeywords = []
    for a in roster:
        allKeywords += a.getKeywords(ao.PROVIDES) + \
                       a.getKeywords(ao.OFFERS)

    # Required keywords.
    for a in roster:

        missing = []
        
        for req in a.getKeywords(ao.REQUIRES):
            if req not in settingKeywords and req not in allKeywords:
                missing.append(req)

        if missing:
            result.append(('missing-requirements', a, missing))

    if result:
        return result        
    
    # Category exclusion.
    #for i in range(rosterSize):
    #    a = roster[i]
    for a in roster:

        exclusions = []
        
        # Does this addon exclude any of the other addons in the
        # roster?
        #for j in range(i + 1, rosterSize):
        #    b = roster[j]
        for b in roster:
            if a is b:
                continue
            
            for category in a.getExcludedCategories():
                if category.isAncestorOf(b.getCategory()):
                    # This is a conflict: a excludes b.
                    exclusions.append(b)
                    break

        if exclusions:
            result.append(('exclusion-by-category', a, exclusions))

    # If category exclusions were found, we won't check any further.
    if result:
        return result

    # Exclusion by keyword.
    for a in roster:

        exclusions = []

        for x in a.getKeywords(ao.EXCLUDES):
            if x in settingKeywords:
                exclusions.append(x)

        if exclusions:
            result.append(('exclusion-by-value', a, exclusions))

        exclusions = []

        # Does this addon exclude any of the other addons in the
        # roster, using keywords?
        for b in roster:
            if a is b:
                continue
            for x in a.getKeywords(ao.EXCLUDES):
                if b.getId() == x or x in b.getKeywords(ao.PROVIDES) or \
                   x in b.getKeywords(ao.OFFERS):
                    exclusions.append((b, x))

        if exclusions:
            result.append(('exclusion-by-keyword', a, exclusions))
        
    # If exclusions were found, we won't check any further.
    if result:
        return result

    # Check for provide conflicts.  Two addons can't provide the
    # keyword.
    for i in range(rosterSize):
        a = roster[i]

        conflicts = []

        for j in range(i + 1, rosterSize):
            b = roster[j]

            for prov in a.getKeywords(ao.PROVIDES):
                if prov in b.getKeywords(ao.PROVIDES):
                    conflicts.append((a, b, prov))

        if conflicts:
            result.append(('provide-conflict', conflicts))

    if result:
        return result

    # Check for offers.  Overrides should be resolved automatically.
    for a in roster:
        overrides = []
        for b in roster:
            if a is b:
                continue
            
            # Does a provide what b only offers?
            for prov in a.getKeywords(ao.PROVIDES):
                if prov in b.getKeywords(ao.OFFERS):
                    if b not in overrides:
                        overrides.append((b, prov))
                    break

        if overrides:
            result.append(('override', a, overrides))

    if result:
        return result

    for i in range(rosterSize):
        a = roster[i]

        overrides = []

        for j in range(i + 1, rosterSize):
            b = roster[j]

            # Does a offer the same thing that b offers?
            for offer in a.getKeywords(ao.OFFERS):
                if offer in b.getKeywords(ao.OFFERS):
                    if b not in overrides:
                        overrides.append((b, offer))

        if overrides:
            result.append(('override', a, overrides))

    if result:
        return result

    # No conflicts found.  Yay!
    return []


def loadAll():
    """Load all addons and manifests."""
    
    # Load all the installed addons in the system and user addon
    # databases.
    for repository in [paths.ADDONS] + paths.getAddonPaths():
        for name in paths.listFiles(repository):
            # Case insensitive.
            if re.search("(?i)^([^.#].*)\.(" + \
                         string.join(ao.EXTENSIONS, '|') + ")$",
                         os.path.basename(name)):
                load(name)

    # Load all the manifests.
    for folder in [paths.getSystemPath(paths.MANIFESTS),
                   paths.getUserPath(paths.MANIFESTS)] + \
                   paths.getAddonPaths():
        for name in paths.listFiles(folder):
            # Case insensitive.
            if paths.hasExtension('manifest', name):
                loadManifest(name)
                
                
def refresh():
    """Empties the data of all loaded addons from the database and reloads
    everything. Sends out a 'addon-database-reloaded' notification."""
    
    init()   
    loadAll()
    events.send(events.Notify('addon-database-reloaded'))
    
    
def init():
    """Initialize the addon database."""
    
    global categories, rootCategory, addons
    
    # Empty the cache.
    categories = []
    addons = {}

    # Create the root category (for unclassified/generic addons).
    rootCategory = Category('')
    categories.append(rootCategory)
    

#
# Module Initialization
#

# Load the metadata cache.
init()
readMetaCache()

loadAll()

# Save the updated cache.
writeMetaCache()



#!/usr/bin/env python

import os, re, subprocess

rootDir = os.path.normpath(os.path.join(os.path.dirname(os.path.abspath(__file__)), '../..'))

def list_projects():
    return subprocess.check_output(['find', '.', '-name', 'project.pbxproj']).split()

print("Generating Xcode projects, hang on...")
for i in range(2):
    os.system("qmake-qt5 -r -spec macx-xcode %s/doomsday.pro CONFIG+='debug deng_notests deng_noccache' > /dev/null 2> /dev/null" % rootDir)
    count = len(list_projects())
    print 'Got %i project files.' % count
    # qmake is quite broken when generating Xcode projects; we'll run again for more success.
    if count > 50: break    
    if i == 0: print "Second time's the charm? Generating again..."

print('Built products will be symlinked in distrib/products/')
for fn in list_projects():
    #print 'patching', fn
    content = file(fn, 'r').read()
    # Symlink products.
    content = re.sub(r'cp -r \$BUILT_PRODUCTS_DIR',
                      'ln -fs $BUILT_PRODUCTS_DIR', content)
    # Quote library paths with space (-L).
    def space_repl(parens):
        def rep(path): return '"\\"' + path.group(1) + '\\""'
        return parens.group(1) + re.sub(r'"([^"]* [^"]*)"', rep, parens.group(3)) + parens.group(4)
    content = re.sub(r'((LIBRARY_SEARCH_PATHS|OTHER_LDFLAGS) = \()([^\)]*)(\);)', space_repl, content)    
    file(fn, 'w').write(content)

print('Setting up a workspace...')
workName = 'Doomsday.xcworkspace'
os.system('mkdir -p %s && cp %s/build/mac/contents.xcworkspacedata %s/' % (workName, rootDir, workName))

print('All done! Open %s in Xcode.' % workName)



# -*- coding: iso-8859-1 -*-
# $Id$
# Snowberry: Extensible Launcher for the Doomsday Engine
#
# Copyright (C) 2004, 2005
#   Jaakko Keränen <jaakko.keranen@iki.fi>
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not: http://www.opensource.org/

## @file host.py Information about the Host Environment
##
## This module contains routines that deal with the host environment
## where Snowberry is running.

import platform
import string
import sys


def isWindows():
    """Check if the host is running Windows.

    @return True, if the operating system is Windows.
    """
    system = platform.system().lower()
    if 'microsoft' in system or 'windows' in system:
        return True
    return False


def isWindowsVista():
    """Check if the host is running Windows Vista or newer."""
    if sys.platform != "win32":
        return False
    winVersion = sys.getwindowsversion()
    return (winVersion[0] >= 6)


def isMac():
    """Check if the host is running a Macintosh operating system.

    @return True, if the operating system is Mac OS X.
    """
    return platform.system() == 'Darwin'


def isUnix():
    """Check if the host is running a variant of Unix.

    @return True, if the operating system is Unix.
    """
    return not isWindows() and not isMac()
    
    
def getEncoding():
    """Returns the character encoding."""
    
    if isMac():
        return "iso-8859-1" #"mac-roman"
    else:
        return "iso-8859-1"



#!/usr/bin/env python
#-*- coding: UTF-8 -*-

"""
This module uses the sample.py script to load all test models it finds.

Note: this is not an exhaustive test suite, it does not check the
data structures in detail. It just verifies whether basic
loading and querying of 3d models using pyassimp works.
"""


import sys,os
import sample
from pyassimp import errors

# paths to be walkd recursively
basepaths = [os.path.join('..','..','..','test','models'), os.path.join('..','..','..','test','models-nonbsd')]

# file extensions to be considered
extensions = ['.3ds','.x','.lwo','.obj','.md5mesh','.dxf','.ply','.stl','.dae','.md5anim','.lws','.irrmesh','.nff','.off','.blend']

def run_tests():
    ok,err = 0,0
    for path in basepaths:
        print("Looking for models in %s..." % path)
        for root, dirs, files in os.walk(path):
            for afile in files:
                base,ext = os.path.splitext(afile)
                if ext in extensions:
                    try:
                        sample.main(os.path.join(root,afile))
                        ok += 1
                    except errors.AssimpError as error:
                        # assimp error is fine, this is a controlled case
                        print error
                        err += 1
                    except Exception:
                        print("Error encountered while loading <%s>"%os.path.join(root,afile))
    print('** Loaded %s models, got controlled errors for %s files' % (ok,err))


if __name__ == '__main__':
	run_tests()







# -*- coding: iso-8859-1 -*-
# $Id$
# Snowberry: Extensible Launcher for the Doomsday Engine
#
# Copyright (C) 2004, 2005
#   Jaakko Keränen <jaakko.keranen@iki.fi>
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not: http://www.opensource.org/

## @file about.py About dialog

import ui, events, language, host
import sb.util.dialog
import widgets as wg
import sb.widget.text as wt
import sb.confdb as st


def init():
    # Listen for the About button.
    events.addCommandListener(handleCommand, ['about', 'show-credits'])
    
    # Commands for the popup menu.
    ui.addMenuCommand(ui.MENU_HELP, 'about')
        

def handleCommand(event):
    """Handle the About command and display the About Snowberry
    dialog.

    @param event  An events.Command event.
    """

    if event.hasId('about'):
        # Create the About dialog and show it.
        dialog, area = sb.util.dialog.createButtonDialog(
            'about-dialog', ['show-credits', '', 'ok'], 'ok', resizable=False)

        content = area.createArea(alignment=ui.ALIGN_VERTICAL, border=0)
        content.setWeight(0)

        # Big logo.
        content.setBorder(16, ui.BORDER_BOTTOM)
        content.createImage('about')

        content.setBorder(6, ui.BORDER_BOTTOM)
        versionText = content.createText('', align=wt.Text.MIDDLE)
        versionText.setBoldStyle()
        versionText.setText(
            language.translate('about-version') + ' ' +
            st.getSystemString('snowberry-version'))
            
        content.createText('about-subtitle', align=wt.Text.MIDDLE)
        content.createText('about-license', align=wt.Text.MIDDLE)
        content.createText('about-website', align=wt.Text.MIDDLE).setSmallStyle()

        #content.setBorder(6)
        #content.setWeight(1)
        #box = content.createArea(boxedWithTitle='about-credits')
        #info = box.createFormattedText()
        #info.setMinSize(300, 280)
        #info.setText(language.translate('about-info'))
    
        dialog.focusWidget('ok')
        dialog.run()
        
    elif event.hasId('show-credits'):
        dialog, area = sb.util.dialog.createButtonDialog(
            'about-credits', ['close'], 'close')
        info = area.createFormattedText()
        info.setMinSize(290, 230)
        info.setText(language.translate('credits'))
        dialog.addEndCommand('close')
        dialog.run()
